import typing
from contextlib import contextmanager

import pytest

from antidote import world
from antidote._internal.argspec import Arguments
from antidote.core.injection import inject, raw_inject, validate_injection
from antidote.exceptions import DependencyNotFoundError


@contextmanager
def does_not_raise():
    yield


class Service:
    pass


class AnotherService:
    pass


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(), does_not_raise()),
    (dict(dependencies="{arg_name}"), does_not_raise()),
    (dict(dependencies=lambda arg: arg.name), does_not_raise()),
    (dict(dependencies=dict(x='x')), does_not_raise()),
    (dict(dependencies=['x']), does_not_raise()),
    (dict(use_names=True), does_not_raise()),
    (dict(use_names=['x']), does_not_raise()),
    (dict(use_type_hints=True), does_not_raise()),
    (dict(use_type_hints=['x']), does_not_raise()),
    (dict(dependencies=1), pytest.raises(TypeError, match=".*dependencies.*int.*")),
    (dict(use_names=1), pytest.raises(TypeError, match=".*use_names.*int.*")),
    (dict(use_type_hints=1), pytest.raises(TypeError, match=".*use_type_hints.*int.*")),
])
def test_validate_injection(kwargs, expectation):
    print(expectation)
    with expectation:
        validate_injection(**kwargs)


@pytest.fixture(params=[raw_inject, inject])
def injector(request):
    return request.param


def test_simple(injector):
    @injector
    def f(x: Service):
        return x

    with world.test.empty():
        s = Service()
        world.singletons.add(Service, s)
        assert s == f()


@pytest.mark.parametrize(
    'expected,kwargs',
    [
        pytest.param((None, None),
                     dict(),
                     id='nothing'),
        pytest.param((Service, None),
                     dict(dependencies=dict(first=Service)),
                     id='dependencies:dict-first'),
        pytest.param((Service, None),
                     dict(dependencies=(Service,)),
                     id='dependencies:tuple-first'),
        pytest.param((None, Service),
                     dict(dependencies=dict(second=Service)),
                     id='dependencies:dict-second'),
        pytest.param((None, Service),
                     dict(dependencies=(None, Service)),
                     id='dependencies:tuple-second'),
        pytest.param(('first', 'second'),
                     dict(dependencies=lambda arg: arg.name),
                     id='dependencies:callable'),
        pytest.param((Service, Service),
                     dict(dependencies=lambda arg: Service),
                     id='dependencies:callable2'),
        pytest.param((None, None),
                     dict(dependencies=lambda s: None),
                     id='dependencies:callable3'),
        pytest.param(('first', 'second'),
                     dict(dependencies="{arg_name}"),
                     id='dependencies:str'),
        pytest.param(('prefix:first', 'prefix:second'),
                     dict(dependencies="prefix:{arg_name}"),
                     id='dependencies:str2'),
        pytest.param(('first', 'second'),
                     dict(use_names=True),
                     id='use_names:True'),
        pytest.param((None, None),
                     dict(use_names=False),
                     id='use_names:False'),
        pytest.param((None, 'second'),
                     dict(use_names=['second']),
                     id='use_names:list')
    ]
)
def test_without_type_hints(injector, expected, kwargs):
    default = object()

    @injector(**kwargs)
    def f(first=default, second=default):
        return first, second

    class A:
        @injector(**kwargs)
        def method(self, first=default, second=default):
            return first, second

        @injector(**kwargs)
        @classmethod
        def class_method(cls, first=default, second=default):
            return first, second

        @injector(**kwargs)
        @staticmethod
        def static_method(first=default, second=default):
            return first, second

        @classmethod
        @injector(**kwargs)
        def klass(cls, first=default, second=default):
            return first, second

    with world.test.empty():
        world.singletons.add({
            Service: Service(),
            AnotherService: AnotherService(),
            'first': object(),
            'second': object(),
            'prefix:first': object(),
            'prefix:second': object()
        })

        expected = tuple((
            world.get(d) if d is not None else default
            for d in expected
        ))
        assert expected == f()
        assert expected == A().method()
        assert expected == A.class_method()
        assert expected == A.static_method()
        assert expected == A.klass()

        a, b = object(), object()
        assert (a, b) == f(a, b)
        assert (a, b) == A().method(a, b)
        assert (a, b) == A.class_method(a, b)
        assert (a, b) == A.static_method(a, b)
        assert (a, b) == A.klass(a, b)


@pytest.mark.parametrize(
    'expected, kwargs',
    [
        pytest.param((Service, None),
                     dict(),
                     id='nothing'),
        pytest.param((Service, None),
                     dict(dependencies=dict(first=Service)),
                     id='dependencies:dict-first'),
        pytest.param((Service, None),
                     dict(dependencies=(Service,)),
                     id='dependencies:tuple-first'),
        pytest.param((Service, Service),
                     dict(dependencies=dict(second=Service)),
                     id='dependencies:dict-second'),
        pytest.param((Service, Service),
                     dict(dependencies=(None, Service)),
                     id='dependencies:tuple-second'),
        pytest.param(('first', 'second'),
                     dict(dependencies=lambda arg: arg.name),
                     id='dependencies:callable'),
        pytest.param((Service, Service),
                     dict(dependencies=lambda arg: Service),
                     id='dependencies:callable2'),
        pytest.param((Service, None),
                     dict(dependencies=lambda arg: None),
                     id='dependencies:callable3'),
        pytest.param(('first', 'second'),
                     dict(dependencies="{arg_name}"),
                     id='dependencies:str'),
        pytest.param(('prefix:first', 'prefix:second'),
                     dict(dependencies="prefix:{arg_name}"),
                     id='dependencies:str2'),
        pytest.param((Service, 'second'),
                     dict(use_names=True),
                     id='use_names:True'),
        pytest.param((Service, None),
                     dict(use_names=False),
                     id='use_names:False'),
        pytest.param((Service, None),
                     dict(use_names=['first']),
                     id='use_names:list-first'),
        pytest.param((Service, 'second'),
                     dict(use_names=['second']),
                     id='use_names:list-second'),
        pytest.param((Service, None),
                     dict(use_type_hints=True),
                     id='use_type_hints:True'),
        pytest.param((Service, None),
                     dict(use_type_hints=['first']),
                     id='use_type_hints:list-first'),
        pytest.param((Service, 'second'),
                     dict(use_type_hints=['first'], use_names=True),
                     id='use_type_hints:list-first+use_names=True'),
        pytest.param((None, None),
                     dict(use_type_hints=['second']),
                     id='use_type_hints:list-second'),
        pytest.param(('first', 'second'),
                     dict(use_type_hints=['second'], use_names=True),
                     id='use_type_hints:list-second+use_names=True'),
        pytest.param((None, None),
                     dict(use_type_hints=False),
                     id='use_type_hints:False'),
        pytest.param(('first', 'second'),
                     dict(use_type_hints=False, use_names=True),
                     id='use_type_hints:False+use_names=True'),
    ]
)
def test_with_type_hints(injector, expected, kwargs):
    default = object()

    @injector(**kwargs)
    def f(first: Service = default, second: str = default):
        return first, second

    class A:
        @injector(**kwargs)
        def method(self, first: Service = default, second: str = default):
            return first, second

        @injector(**kwargs)
        @classmethod
        def class_method(cls, first: Service = default, second: str = default):
            return first, second

        @injector(**kwargs)
        @staticmethod
        def static_method(first: Service = default, second: str = default):
            return first, second

        @classmethod
        @injector(**kwargs)
        def klass(cls, first: Service = default, second: str = default):
            return first, second

    with world.test.empty():
        world.singletons.add({Service: Service(),
                              AnotherService: AnotherService(),
                              'first': object(),
                              'second': object(),
                              'prefix:first': object(),
                              'prefix:second': object()})

        expected = tuple((
            world.get(d) if d is not None else default
            for d in expected
        ))
        assert expected == f()
        assert expected == A().method()
        assert expected == A.class_method()
        assert expected == A.static_method()
        assert expected == A.klass()

        a, b = object(), object()
        assert (a, b) == f(a, b)
        assert (a, b) == A().method(a, b)
        assert (a, b) == A.class_method(a, b)
        assert (a, b) == A.static_method(a, b)
        assert (a, b) == A.klass(a, b)


@pytest.mark.parametrize('type_hint',
                         [str, int, float, set, list, dict, complex, type, tuple, bytes,
                          bytearray, typing.Optional, typing.Sequence])
def test_ignored_type_hints(injector, type_hint):
    @injector
    def f(x: type_hint):
        pass

    with world.test.empty():
        world.singletons.add(type_hint, object())
        with pytest.raises(TypeError):
            f()


def test_arguments():
    def f(a, b):
        pass

    arguments = Arguments.from_callable(f)

    @raw_inject(arguments=arguments, use_names=True)
    def g(**kwargs):
        return kwargs

    with world.test.empty():
        world.singletons.add(dict(a=12, b=24))
        assert dict(a=12, b=24) == g()


def test_none_optional_support(injector):
    class Dummy:
        pass

    @injector
    def f(x: Service = None):
        return x

    @injector
    def g(x: typing.Optional[Service] = None):
        return x

    @injector
    def h(x: Dummy, y: typing.Optional[Service] = None):
        return y

    @injector
    def f2(x: typing.Union[Service, Dummy]):
        pass

    with world.test.empty():
        s = Service()
        world.singletons.add(Dummy, Dummy())
        world.singletons.add(Service, s)
        assert s == f()
        assert s == g()
        assert s == h()
        with pytest.raises(TypeError):
            f2()

    with world.test.empty():
        world.singletons.add(Dummy, Dummy())
        assert f() is None
        assert g() is None
        assert h() is None


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(TypeError, match=".*dependencies.*"),
                     dict(dependencies=object()),
                     id="dependencies:unsupported-type"),
        pytest.param(pytest.raises(ValueError, match=".*dependencies.*"),
                     dict(dependencies="dummy"),
                     id="dependencies:missing-string-arg_name"),
        pytest.param(pytest.raises(TypeError),
                     dict(dependencies={1: 'x'}),
                     id="dependencies:invalid-key-type"),
        pytest.param(pytest.raises(TypeError, match=".*use_names.*"),
                     dict(use_names=object()),
                     id="use_names:unsupported-type"),
        pytest.param(pytest.raises(TypeError),
                     dict(use_names=[1]),
                     id="use_names:invalid-name-type"),
        pytest.param(pytest.raises(TypeError, match=".*use_type_hints.*"),
                     dict(use_type_hints=object()),
                     id="use_type_hints:unsupported-type"),
        pytest.param(pytest.raises(TypeError),
                     dict(use_type_hints=[1]),
                     id="use_type_hints:invalid-name-type")
    ]
)
def test_invalid_args(injector, expectation, kwargs):
    with expectation:
        @injector(**kwargs)
        def f(x):
            return x

        f()

    with expectation:
        class A:
            @injector(**kwargs)
            def method(self, x):
                return x

        A().method()

    with expectation:
        class A:
            @injector(**kwargs)
            @classmethod
            def classmethod(cls, x):
                return x

        A.classmethod()

    with expectation:
        class A:
            @injector(**kwargs)
            @staticmethod
            def staticmethod(x):
                return x

        A.staticmethod()

    with expectation:
        class A:
            @classmethod
            @injector(**kwargs)
            def classmethod(cls, x):
                return x

        A.classmethod()

    with expectation:
        validate_injection(**kwargs)


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(TypeError),
                     dict(),
                     id="unknown-dependency"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(dependencies=(Service,)),
                     id="dependencies:unknown-dependency-tuple"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(dependencies=dict(x=Service)),
                     id="dependencies:unknown-dependency-dict"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*Service.*"),
                     dict(dependencies=lambda arg: Service),
                     id="dependencies:unknown-dependency-callable"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*unknown.*"),
                     dict(dependencies="unknown:{arg_name}"),
                     id="dependencies:unknown-dependency-str"),
        pytest.param(pytest.raises((ValueError, TypeError)),
                     dict(dependencies=(None, None)),
                     id="dependencies:too-much-arguments"),
        pytest.param(pytest.raises(ValueError, match=".*unknown.*"),
                     dict(dependencies=dict(unknown=DependencyNotFoundError)),
                     id="dependencies:unknown-argument-dict"),
        pytest.param(pytest.raises(TypeError),
                     dict(use_names=False),
                     id="use_names:unknown-dependency-False"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*x.*"),
                     dict(use_names=True),
                     id="use_names:unknown-dependency-True"),
        pytest.param(pytest.raises(DependencyNotFoundError, match=".*x.*"),
                     dict(use_names=['x']),
                     id="use_names:unknown-dependency-list"),
        pytest.param(pytest.raises(ValueError, match=".*y.*"),
                     dict(use_names=['y']),
                     id="use_names:unknown-argument-list"),
        pytest.param(pytest.raises(ValueError, match=".*y.*"),
                     dict(use_names=['x', 'y']),
                     id="use_names:unknown-argument-list2"),
        pytest.param(pytest.raises(TypeError),
                     dict(use_names=[]),
                     id="use_names:empty"),
        pytest.param(pytest.raises(ValueError, match=".*y.*"),
                     dict(use_type_hints=['y']),
                     id="use_type_hints:unknown-arg"),
    ]
)
def test_invalid_call(injector, expectation, kwargs):
    with expectation:
        @injector(**kwargs)
        def f(x):
            return x

        f()

    with expectation:
        class A:
            @injector(**kwargs)
            def method(self, x):
                return x

        A().method()

    with expectation:
        class A:
            @injector(**kwargs)
            @classmethod
            def classmethod(cls, x):
                return x

        A.classmethod()

    with expectation:
        class A:
            @injector(**kwargs)
            @staticmethod
            def staticmethod(x):
                return x

        A.staticmethod()

    with expectation:
        class A:
            @classmethod
            @injector(**kwargs)
            def classmethod(cls, x):
                return x

        A.classmethod()


@pytest.mark.parametrize(
    'expectation,kwargs',
    [
        pytest.param(pytest.raises(ValueError, match=".*self.*"),
                     dict(dependencies=dict(self='x')),
                     id="dependencies"),
        pytest.param(pytest.raises(ValueError, match=".*self.*"),
                     dict(use_names=('self',)),
                     id="use_names"),
        pytest.param(pytest.raises(ValueError, match=".*self.*"),
                     dict(use_type_hints=('self',)),
                     id="use_type_hints"),
    ]
)
def test_cannot_inject_self(injector, expectation, kwargs):
    with expectation:
        class A:
            @injector(**kwargs)
            def method(self, x=None):
                return x

        A()

    with expectation:
        class A:
            @injector(**kwargs)
            @classmethod
            def classmethod(self, x=None):
                return x

        A()

    with expectation:
        class A:
            @classmethod
            @injector(**kwargs)
            def classmethod(self, x=None):
                return x

        A()


def test_invalid_type_hint(injector):
    @injector
    def f(x: Service):
        return x

    with pytest.raises(DependencyNotFoundError):
        f()


def test_no_injections():
    def f(x):
        return x

    # When nothing can be injected, the same function should be returned
    assert raw_inject(f) is f


def test_already_injected(injector):
    @injector(use_names=True)
    def f(x):
        return x

    # When the function has already its arguments injected, the same function should
    # be returned
    assert raw_inject(f) is f


def test_invalid_inject(injector):
    with pytest.raises(TypeError):
        @injector
        class Dummy:
            pass

    with pytest.raises(TypeError):
        injector(1)



"""
Cython version of the wrapper, doing the same thing but faster.
"""
# @formatter:off
cimport cython
from cpython.dict cimport PyDict_Copy, PyDict_New
from cpython.object cimport PyObject_Call, PyObject_CallMethodObjArgs
from cpython.ref cimport PyObject

from antidote._internal.state cimport fast_get_container
from antidote.core.container cimport DependencyResult, PyObjectBox, RawContainer
from ..core.exceptions import DependencyNotFoundError

# @formatter:on

cdef extern from "Python.h":
    PyObject*PyTuple_GET_ITEM(PyObject *p, Py_ssize_t pos)
    Py_ssize_t PyTuple_GET_SIZE(PyObject *p)
    int PyDict_Contains(PyObject *p, PyObject *key) except -1
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1


compiled = True

cdef class Injection:
    cdef:
        str arg_name
        bint required
        object dependency

    def __repr__(self):
        return f"{type(self).__name__}(arg_name={self.arg_name!r}, " \
               f"required={self.required!r}, dependency={self.dependency!r})"

    def __init__(self, str arg_name, bint required, object dependency):
        self.arg_name = arg_name
        self.required = required
        self.dependency = dependency

cdef class InjectionBlueprint:
    cdef:
        tuple injections

    def __init__(self, tuple injections):
        self.injections = injections

    def is_empty(self):
        cdef:
            Injection injection
        for injection in self.injections:
            if injection.dependency is not None:
                return False
        return True

def build_wrapper(InjectionBlueprint blueprint,
                  object wrapped,
                  bint skip_first = False):
    """
    Not relying on __init__ neither __cinit__ makes initialization faster which
    is especially important during __get__ as a InjectedBoundWrapper needs to
    be created *every* time.
    """
    cdef:
        InjectedWrapper wrapper = InjectedWrapper.__new__(InjectedWrapper)
    wrapper.__wrapped__ = wrapped
    wrapper.__blueprint = blueprint
    wrapper.__injection_offset = 1 if skip_first else 0
    wrapper.__is_classmethod = isinstance(wrapped, classmethod)
    wrapper.__is_staticmethod = isinstance(wrapped, staticmethod)
    return wrapper

def get_wrapper_dependencies(wrapper):
    if not isinstance(wrapper, InjectedWrapper):
        raise TypeError(f"Argument must be an {InjectedWrapper}")

    return (<InjectedWrapper> wrapper).get_injections()

def is_wrapper(x) -> bool:
    return isinstance(x, InjectedWrapper)

@cython.freelist(128)
cdef class InjectedWrapper:
    cdef:
        dict __dict__
        readonly object __wrapped__
        InjectionBlueprint __blueprint
        int __injection_offset
        bint __is_classmethod
        bint __is_staticmethod

    cdef list get_injections(self):
        cdef:
            Injection inj
        return [inj.dependency
                for inj in self.__blueprint.injections
                if inj.dependency is not None]

    def __call__(self, *args, **kwargs):
        cdef:
            RawContainer container = fast_get_container()
            DependencyResult result
            PyObjectBox box = PyObjectBox.__new__(PyObjectBox)
            PyObject*injection
            PyObject*arg_name
            PyObject*injections = <PyObject*> self.__blueprint.injections
            bint dirty_kwargs = False
            Py_ssize_t i
            Py_ssize_t offset = self.__injection_offset + PyTuple_GET_SIZE(
                <PyObject*> args)
            Py_ssize_t n = PyTuple_GET_SIZE(injections)
        result.box = <PyObject*> box

        if kwargs:
            for i in range(offset, n):
                injection = PyTuple_GET_ITEM(injections, i)
                if (<Injection> injection).dependency is not None:
                    arg_name = <PyObject*> (<Injection> injection).arg_name
                    if PyDict_Contains(<PyObject*> kwargs, arg_name) == 0:
                        container.fast_get(
                            <PyObject*> (<Injection> injection).dependency, &result)
                        if result.flags != 0:
                            if not dirty_kwargs:
                                kwargs = PyDict_Copy(kwargs)
                                dirty_kwargs = True
                            PyDict_SetItem(<PyObject*> kwargs,
                                           arg_name,
                                           <PyObject*> box.obj)
                        elif (<Injection> injection).required:
                            raise DependencyNotFoundError(
                                (<Injection> injection).dependency)
        else:
            for i in range(offset, n):
                injection = PyTuple_GET_ITEM(injections, i)
                if (<Injection> injection).dependency is not None:
                    container.fast_get(<PyObject*> (<Injection> injection).dependency,
                                       &result)
                    if result.flags != 0:
                        if not dirty_kwargs:
                            kwargs = PyDict_New()
                            dirty_kwargs = True
                        PyDict_SetItem(
                            <PyObject*> kwargs,
                            <PyObject*> (<Injection> injection).arg_name,
                            <PyObject*> box.obj
                        )
                    elif (<Injection> injection).required:
                        raise DependencyNotFoundError((<Injection> injection).dependency)

        return PyObject_Call(self.__wrapped__, args, kwargs)

    def __get__(self, instance, owner):
        cdef:
            InjectedBoundWrapper wrapper = InjectedBoundWrapper.__new__(
                InjectedBoundWrapper)
        wrapper.__wrapped__ = PyObject_CallMethodObjArgs(self.__wrapped__,
                                                         "__get__",
                                                         <PyObject*> instance,
                                                         <PyObject*> owner,
                                                         NULL)
        wrapper.__blueprint = self.__blueprint
        if self.__is_classmethod or (not self.__is_staticmethod and instance is not None):
            wrapper.__injection_offset = 1
        else:
            wrapper.__injection_offset = 0
        wrapper.__is_classmethod = False
        wrapper.__is_staticmethod = False

        return wrapper

    def __getattr__(self, name):
        return getattr(self.__wrapped__, name)

cdef class InjectedBoundWrapper(InjectedWrapper):
    def __get__(self, instance, owner):
        return self



import collections.abc as c_abc
import enum
import inspect
from typing import Callable, FrozenSet, Iterable, Optional, overload, TypeVar, Union

from .injection import DEPENDENCIES_TYPE, raw_inject, validate_injection
from .._compatibility.typing import final
from .._internal import API
from .._internal.argspec import Arguments
from .._internal.utils import Copy, FinalImmutable, raw_getattr

C = TypeVar('C', bound=type)
_empty_set: FrozenSet[str] = frozenset()


@API.public
@final
class Wiring(FinalImmutable):
    """
    Defines how a class should be wired, meaning if/how/which methods are injected. This
    class is intended to be used by configuration objects. If you just want to wire a
    single class, consider using the class decorator :py:func:`~.wire` instead.

    Injection arguments (:code:`dependencies`, :code:`use_names`,  :code:`use_type_hints`)
    are adapted for match the arguments of the method. Hence :py:func:`~.injection.inject`
    won't raise an error that it has too much dependencies.

    Instances are immutable. If you want to change some parameters, typically defaults
    defined by Antidote, you'll need to rely on :py:meth:`~.copy`. It accepts the same
    arguments as :py:meth:`~.__init__` and overrides existing values.

    .. doctest:: core_Wiring

        >>> from antidote import Wiring
        >>> # Methods must always be specified.
        ... w = Wiring(methods=['my_method', 'other_method'])
        >>> # Now argument names will be used on both my_method and other_method.
        ... w_copy = w.copy(use_names=True)

    """
    __slots__ = ('methods', 'dependencies', 'use_names', 'use_type_hints', 'wire_super',
                 'ignore_missing_method')
    methods: FrozenSet[str]
    """Method names that must be injected."""
    wire_super: FrozenSet[str]
    """
    Method names which may be retrieved in a parent class. By default, only methods
    directly defined in the class itself may be injected.
    """
    ignore_missing_method: FrozenSet[str]
    """Method names for which an exception will not be raised if not found."""
    dependencies: DEPENDENCIES_TYPE
    use_names: Union[bool, FrozenSet[str]]
    use_type_hints: Union[bool, FrozenSet[str]]

    def __init__(self,
                 *,
                 methods: Iterable[str],
                 dependencies: DEPENDENCIES_TYPE = None,
                 use_names: Union[bool, Iterable[str]] = None,
                 use_type_hints: Union[bool, Iterable[str]] = None,
                 wire_super: Union[bool, Iterable[str]] = None,
                 ignore_missing_method: Union[bool, Iterable[str]] = None):
        """
        Args:
            methods: Methods names.
            dependencies: Propagated for every method to :py:func:`~.injection.inject`
            use_names: Propagated for every method to :py:func:`~.injection.inject`
            use_type_hints: Propagated for every method to :py:func:`~.injection.inject`
            wire_super: Method names which may be retrieved in a parent class. By default,
                only methods  directly defined in the class itself may be injected. If
                :py:obj:`True` all methods will be searched in parent classes. Defaults to
                :py:obj:`False`.
            ignore_missing_method:  Method names for which an exception will not be
                raised if not found. If :py:obj:`True`, no exception will be raised if
                any of the methods is missing. Defaults to :py:obj:`False`.
        """
        if not isinstance(methods, c_abc.Iterable) or isinstance(methods, str):
            raise TypeError(f"Methods must be an iterable of method names, "
                            f"not {type(methods)}.")
        else:
            methods = frozenset(methods)
        if len(methods) == 0:
            raise ValueError("At least one method must be wired")
        if not all(isinstance(method, str) for method in methods):
            raise ValueError("Methods are expected to be names (str)")

        if not isinstance(use_names, str) and isinstance(use_names, c_abc.Iterable):
            use_names = frozenset(use_names)
        if not isinstance(use_type_hints, str) and isinstance(use_type_hints,
                                                              c_abc.Iterable):
            use_type_hints = frozenset(use_type_hints)
        validate_injection(dependencies, use_names, use_type_hints)

        if not (wire_super is None or isinstance(wire_super, (bool, c_abc.Iterable))) \
                or isinstance(wire_super, str):
            raise TypeError(f"wire_super must be either a boolean or a whitelist of "
                            f"methods names, not {type(wire_super)!r}.")
        if wire_super is None or isinstance(wire_super, bool):
            wire_super = bool(wire_super)
        else:
            wire_super = frozenset(wire_super)
            if not wire_super.issubset(methods):
                raise ValueError(f"wire_super is not a subset of methods. "
                                 f"Method names {wire_super - methods!r} are unknown.")

        if isinstance(ignore_missing_method, str) \
                or not (ignore_missing_method is None
                        or isinstance(ignore_missing_method, (bool, c_abc.Iterable))):
            raise TypeError(f"ignore_missing_method must be a boolean or a list of "
                            f"methods names for which an exception must be raised if "
                            f"it's missing, not {type(ignore_missing_method)}")
        if ignore_missing_method is None or isinstance(ignore_missing_method, bool):
            ignore_missing_method = bool(ignore_missing_method)
        else:
            ignore_missing_method = frozenset(ignore_missing_method)
            if not ignore_missing_method.issubset(methods):
                unexpected = ignore_missing_method - methods
                raise ValueError(f"ignore_missing_method is not a subset of methods. "
                                 f"Method names {', '.join(map(repr, unexpected))} "
                                 f"are unknown.")

        if isinstance(wire_super, bool):
            wire_super = methods if wire_super else _empty_set
        if isinstance(ignore_missing_method, bool):
            ignore_missing_method = methods if ignore_missing_method else _empty_set

        super().__init__(methods=methods,
                         wire_super=wire_super,
                         dependencies=dependencies,
                         use_names=use_names,
                         use_type_hints=use_type_hints,
                         ignore_missing_method=ignore_missing_method)

    def copy(self,
             *,
             methods: Union[Iterable[str], Copy] = Copy.IDENTICAL,
             dependencies: Union[DEPENDENCIES_TYPE, Copy] = Copy.IDENTICAL,
             use_names: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL,
             use_type_hints: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL,
             wire_super: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL,
             ignore_missing_method: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL
             ) -> 'Wiring':
        """
        Returns:
            Wiring: copy of the current object with the specified attributes overriding
            existing ones.
        """
        return Wiring(
            methods=self.methods if methods is Copy.IDENTICAL else methods,
            dependencies=(self.dependencies
                          if dependencies is Copy.IDENTICAL else
                          dependencies),
            use_names=self.use_names if use_names is Copy.IDENTICAL else use_names,
            use_type_hints=(self.use_type_hints
                            if use_type_hints is Copy.IDENTICAL else
                            use_type_hints),
            wire_super=self.wire_super if wire_super is Copy.IDENTICAL else wire_super,
            ignore_missing_method=(self.ignore_missing_method
                                   if ignore_missing_method is Copy.IDENTICAL else
                                   ignore_missing_method)
        )

    @API.experimental
    def wire(self, klass: C) -> C:
        """**Experimental**

        Used to wire a class with specified configuration.

        Args:
            klass: Class to wired

        Returns:
            The same class object with specified methods injected.
        """
        return _wire_class(klass, self)


@overload
def wire(klass: C,  # noqa: E704  # pragma: no cover
         *,
         methods: Iterable[str],
         dependencies: DEPENDENCIES_TYPE = None,
         use_names: Union[bool, Iterable[str]] = None,
         use_type_hints: Union[bool, Iterable[str]] = None,
         wire_super: Union[bool, Iterable[str]] = None
         ) -> C: ...


@overload
def wire(*,  # noqa: E704  # pragma: no cover
         methods: Iterable[str],
         dependencies: DEPENDENCIES_TYPE = None,
         use_names: Union[bool, Iterable[str]] = None,
         use_type_hints: Union[bool, Iterable[str]] = None,
         wire_super: Union[bool, Iterable[str]] = None
         ) -> Callable[[C], C]: ...


@API.public
def wire(klass: type = None,
         *,
         methods: Iterable[str],
         dependencies: DEPENDENCIES_TYPE = None,
         use_names: Union[bool, Iterable[str]] = None,
         use_type_hints: Union[bool, Iterable[str]] = None,
         wire_super: Union[bool, Iterable[str]] = None):
    """Wire a class by injecting specified methods.

    Injection arguments (:code:`dependencies`, :code:`use_names`,  :code:`use_type_hints`)
    are adapted for match the arguments of the method. Hence :py:func:`~.injection.inject`
    won't raise an error that it has too much dependencies.

    Args:
        klass: Class to wire.
        methods: Method names that must be injected.
        dependencies: Propagated for every method to :py:func:`~.injection.inject`.
        use_names: Propagated for every method to :py:func:`~.injection.inject`.
        use_type_hints: Propagated for every method to :py:func:`~.injection.inject`.
        wire_super: Method names which may be retrieved in a parent class. By default,
            only methods directly defined in the class itself may be injected. If set to
            :py:obj:`True`, it will applied to all methods..

    Returns:
        Wired class or a class decorator.

    """
    wiring = Wiring(
        methods=methods,
        wire_super=wire_super,
        dependencies=dependencies,
        use_names=use_names,
        use_type_hints=use_type_hints
    )

    def wire_methods(cls):
        return _wire_class(cls, wiring)

    return klass and wire_methods(klass) or wire_methods


W = TypeVar('W', bound='WithWiringMixin')


@API.experimental
class WithWiringMixin:
    """**Experimental**

    Used by configuration classes (immutable having a :code:`copy()` method) with a
    :code:`wiring` attribute to change it more simply with the :py:meth:`~.with_wiring`
    method.
    """
    __slots__ = ()

    wiring: Optional[Wiring]

    def copy(self: W, *, wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL) -> W:
        raise NotImplementedError()  # pragma: no cover

    def with_wiring(self: W,
                    *,
                    methods: Union[Iterable[str], Copy] = Copy.IDENTICAL,
                    dependencies: Union[DEPENDENCIES_TYPE, Copy] = Copy.IDENTICAL,
                    use_names: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL,
                    use_type_hints: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL,
                    wire_super: Union[bool, Iterable[str], Copy] = Copy.IDENTICAL,
                    ignore_missing_method: Union[
                        bool, Iterable[str], Copy] = Copy.IDENTICAL
                    ) -> W:
        """
        Accepts the same arguments as :py:class:`~.Wiring`. And behaves the same way
        as :py:meth:`.Wiring.copy`.

        Returns:
            Copy of current instance with its :code:`wiring` attribute modified with
            provided arguments.
        """
        if self.wiring is None:
            if methods is Copy.IDENTICAL:
                raise TypeError("Current wiring is None, so methods must be specified")
            wiring = Wiring(methods=methods)
            return self.copy(wiring=wiring.copy(
                dependencies=dependencies,
                use_names=use_names,
                use_type_hints=use_type_hints,
                wire_super=wire_super,
                ignore_missing_method=ignore_missing_method))
        else:
            return self.copy(wiring=self.wiring.copy(
                methods=methods,
                dependencies=dependencies,
                use_names=use_names,
                use_type_hints=use_type_hints,
                wire_super=wire_super,
                ignore_missing_method=ignore_missing_method))


@API.private  # Used internally for auto wiring.
class AutoWire(enum.Enum):
    auto = object()


@API.private  # Use wire() or Wiring.wire() instead
def _wire_class(cls: C, wiring: Wiring) -> C:
    if not inspect.isclass(cls):
        raise TypeError(f"Expecting a class, got a {type(cls)}")

    for method_name in wiring.methods:
        try:
            method = raw_getattr(cls, method_name,
                                 with_super=method_name in wiring.wire_super)
        except AttributeError:
            if method_name not in wiring.ignore_missing_method:
                raise
        else:
            arguments = Arguments.from_callable(method)
            use_names = wiring.use_names
            use_type_hints = wiring.use_type_hints
            dependencies = wiring.dependencies

            # Restrict injection parameters to those really needed by the method.
            if isinstance(dependencies, c_abc.Mapping):
                dependencies = {
                    arg_name: dependency
                    for arg_name, dependency in dependencies.items()
                    if arg_name in arguments.without_self
                }
            elif isinstance(dependencies, c_abc.Sequence) \
                    and not isinstance(dependencies, str):
                dependencies = dependencies[:len(arguments.without_self)]

            if use_names is not None and not isinstance(use_names, bool):
                use_names = use_names.intersection(arguments.arg_names)

            if use_type_hints is not None and not isinstance(use_type_hints, bool):
                use_type_hints = use_type_hints.intersection(arguments.arg_names)

            # If we're wrapping a static/class-method, we can just re-wrap it so
            # that isinstance(..., classmethod) still works as one would expect.
            injected_method = raw_inject(
                method.__func__
                if isinstance(method, (classmethod, staticmethod)) else
                method,
                arguments=arguments,
                dependencies=dependencies,
                use_names=use_names,
                use_type_hints=use_type_hints
            )

            if isinstance(method, classmethod):
                injected_method = classmethod(injected_method)
            elif isinstance(method, staticmethod):
                injected_method = staticmethod(injected_method)

            if injected_method is not method:  # If something has changed
                setattr(cls, method_name, injected_method)

    return cls



from antidote import const, Constants, world


def test_run_me() -> None:
    with world.test.empty():
        world.singletons.add('test', [])
        world.get[list]('test').append(1)
        world.lazy[list]('test').get().append(2)

        world.get('test').append(1)

        class Conf(Constants):
            A = const[list]("a")

            def get(self, key):
                return []

        Conf().A.append(1)



from typing import Any, cast, Generic, Hashable, Sequence, TypeVar

from .._compatibility.typing import final
from .._internal import API
from .._internal.utils import FinalImmutable, Immutable
from .._internal.utils.debug import debug_repr
from .._internal.utils.immutable import ImmutableGenericMeta

T = TypeVar('T')


@API.public
@final
class Dependency(Immutable, Generic[T], metaclass=ImmutableGenericMeta):
    """
    Used to clearly state that a value should be treated as a dependency and must
    be retrieved from Antidote. It is recommended to use it through
    :py:func:`~antidote.world.lazy` as presented:

    .. doctest:: core_Dependency

        >>> from antidote import world
        >>> world.singletons.add('dependency', 1)
        >>> world.lazy('dependency')
        Dependency(value='dependency')
        >>> # to retrieve the dependency later, you may use get()
        ... world.lazy[int]('dependency').get()
        1

    """
    __slots__ = ('value',)
    value: Hashable
    """Dependency to be retrieved"""

    def __init__(self, value: Hashable):
        """
        Args:
            value: actual dependency to be retrieved later.
        """
        super().__init__(value=value)

    def get(self) -> T:
        """
        Returns:
            dependency instance retrieved from :py:mod:`~antidote.world`.
        """
        from antidote import world
        return cast(T, world.get(self.value))

    @API.private
    def __antidote_debug_repr__(self) -> str:
        return f"Dependency(value={debug_repr(self.value)})"


@API.public
@final
class DependencyDebug(FinalImmutable):
    """
    Debug information on a dependency. Used by :py:mod:`.world.debug` to provide runtime
    information for debugging.
    """
    __slots__ = ('info', 'singleton', 'wired', 'dependencies')
    info: str
    singleton: bool
    wired: Sequence
    dependencies: Sequence[Hashable]

    def __init__(self,
                 info: str,
                 *,
                 singleton: bool,
                 wired: Sequence = tuple(),
                 dependencies: Sequence[Hashable] = tuple()):
        """
        Args:
            info: Short and concise information on the dependency, just enough to identify
                clearly which one it is.
            singleton: Whether the dependency is a singleton or not.
            wired: Every class or function that may have been wired for this dependency.
            dependencies: Any transient dependency, so dependencies of this dependency.
        """
        super().__init__(info, singleton, wired, dependencies)

    def __eq__(self, other):
        return isinstance(other, DependencyDebug) \
               and self.info == other.info \
               and self.singleton == other.singleton \
               and self.wired == other.wired \
               and self.dependencies == other.dependencies  # noqa: E126



import collections.abc as c_abc
from typing import Callable, cast, FrozenSet, Iterable, Optional, Sequence, Union

from ._compatibility.typing import final
from ._constants import ConstantsMeta, MakeConst
from ._internal import API
from ._internal.utils import Copy, FinalImmutable
from .core.wiring import Wiring, WithWiringMixin

const = MakeConst()


@API.public
class Constants(metaclass=ConstantsMeta, abstract=True):
    """
    Used to declare constants, typically configuration. The purpose is to make common
    configuration manipulation easy. If you need complex operations, consider using the
    Constants class as a service by specifying :code:`public=True`.

    By default all public uppercase attributes are considered to be constants. When using
    them as dependencies or directly through a instance the method :code:`get()` will be
    called with its associated value. Only :code:`__init__()` is wired by default.

    The purpose is to have easily indentify where a specific constant is used in your
    code while keeping it lazy. The class and the constants themselves will only be
    instantiated if necessary.

    .. doctest:: helpers_Constants

        >>> from antidote import Constants, world
        >>> class Config(Constants):
        ...     # By default only public uppercase attributes are considered constants
        ...     DOMAIN = 'domain'
        ...     _A = 'unchanged'
        ...     a = 'unchanged'
        ...
        ...     def __init__(self):
        ...         self._data = {'domain': 'example.com'}
        ...
        ...     # By default the method get() is used to retrieve the constant.
        ...     def get(self, key):
        ...         return self._data[key]
        ...
        >>> Config._A
        'unchanged'
        >>> Config.a
        'unchanged'
        >>> world.get(Config.DOMAIN)
        'example.com'
        >>> # For ease of use, if accessed through an instance, the constant will not
        ... # pass through Antidote.
        ... Config().DOMAIN
        'example.com'

    You may also use :py:func:`.const` if you want to explitly define it as a constant or
    if you want to specify its type. By default if the type is one of :code:`str`,
    :code:`float` or :code:`int`, it will be cast automatically. You can control this
    behavior with the :code:`auto_cast` argument of :py:attr:`~.Constants.Conf`.

    .. doctest:: helpers_Constants_v2

        >>> from antidote import Constants, const, world
        >>> class Config(Constants):
        ...     # As `post` is not uppercase, it would not have been considered to
        ...     # be a constant by default.
        ...     port = const[int]('80')
        ...     port2 = const('8080')
        ...     dummy = const[dict]('dummy')
        ...
        ...     def get(self, value):
        ...         return value
        >>> # Mypy will treat it as a int AND it'll be cast to int.
        ... Config().port
        80
        >>> world.get(Config.port)
        80
        >>> Config().port2
        '8080'
        >>> world.get(Config.port2)
        '8080'
        >>> # Mypy will treat it as a dict but it'll not be cast to anything
        ... Config().dummy
        'dummy'

    You may customize how the Constants class is configured through
    :py:attr:`.__antidote__`.  Among others you can also customize
    :py:attr:`~.Constants.Conf.is_const` used to determine which attributes are constants
    or not.

    .. doctest:: helpers_Constants_v3

        >>> from antidote import Constants, const, world
        >>> class Config(Constants):
        ...     __antidote__ = Constants.Conf(is_const=lambda name: name.startswith("CONF_"))
        ...     CONF_A = "A"
        ...     A = "A"
        ...
        ...     def get(self, value):
        ...         return f"Hello {value}"
        >>> Config().CONF_A
        'Hello A'

    """  # noqa: E501

    @final
    class Conf(FinalImmutable, WithWiringMixin):
        """
        Immutable constants configuration. To change parameters on a existing instance,
        use either method :py:meth:`.copy` or
        :py:meth:`~.core.wiring.WithWiringMixin.with_wiring`.
        """
        __slots__ = ('wiring', 'is_const', 'public', 'auto_cast')
        wiring: Optional[Wiring]
        is_const: Callable[[str], bool]
        public: bool
        auto_cast: FrozenSet[type]

        def __init__(self,
                     *,
                     public: bool = False,
                     is_const: Optional[Callable[[str], bool]] =
                     lambda name: name.isupper() and not name.startswith('_'),
                     auto_cast: Union[Iterable[type], bool] = True,
                     wiring: Optional[Wiring] = Wiring(methods=['__init__'],
                                                       ignore_missing_method=[
                                                           '__init__'])):
            """
            Args:
                public: Whether the Constants instance is available through Antidote like
                    a service or not. Defaults to :py:obj:`False`.
                is_const: Callable determining whether attributes are considered to be
                    constant or not. If :py:obj:`None`, no attributes will be changed.
                    Defaults to changing all public (not starting with an underscore)
                    uppercase attributes.
                wiring: :py:class:`Wiring` used on the class. Defaults to wire only
                    :code:`__init__()`.
                auto_cast: When the type of the constant is specified with
                    :py:func:`.const`, the value of the dependency will be cast to its
                    type if it's one of :code:`str`, :code:`float` or :code:`int` by
                    default. You can disable this by specifying code:`auto_cast=False` or
                    change the types for which it's done by specifying explicitly those
                    types.
            """
            if not isinstance(public, bool):
                raise TypeError(f"public must be a boolean, "
                                f"not a {type(public)}")

            if not (wiring is None or isinstance(wiring, Wiring)):
                raise TypeError(f"wiring can be None or a Wiring, "
                                f"but not a {type(wiring)}")

            if isinstance(auto_cast, bool):
                if auto_cast:
                    auto_cast = frozenset((float, int, str))
                else:
                    auto_cast = frozenset()
            elif isinstance(auto_cast, c_abc.Iterable):
                auto_cast = frozenset(auto_cast)
                if not all(isinstance(tpe, type) for tpe in auto_cast):
                    raise TypeError("auto_cast must a boolean or a iterable of types")
            else:
                raise TypeError("auto_cast must a boolean or a iterable of types")

            if is_const is None:
                def is_const(name):
                    return False

            if not callable(is_const):
                raise TypeError(f"is_const must be a callable, not a {type(is_const)}")
            super().__init__(wiring=wiring, is_const=is_const, public=public,
                             auto_cast=auto_cast)

        def copy(self,
                 *,
                 public: Union[bool, Copy] = Copy.IDENTICAL,
                 wiring: Union[Optional[Wiring], Copy] = Copy.IDENTICAL,
                 auto_cast: Union[Union[Sequence[type], bool], Copy] = Copy.IDENTICAL,
                 is_const: Union[Optional[Callable[[str], bool]], Copy] = Copy.IDENTICAL):

            return Constants.Conf(
                public=self.public if public is Copy.IDENTICAL else public,
                wiring=self.wiring if wiring is Copy.IDENTICAL else wiring,
                auto_cast=self.auto_cast if auto_cast is Copy.IDENTICAL else auto_cast,
                # Waiting for a fix: https://github.com/python/mypy/issues/6910
                is_const=(cast(Callable[[str], bool], getattr(self, 'is_const'))
                          if is_const is Copy.IDENTICAL else
                          is_const)
            )

    __antidote__: Conf = Conf()



from typing import cast

import pytest

from antidote import (Factory, factory, Service, Tag, Tagged,
                      world)


@pytest.fixture(autouse=True)
def new_world():
    with world.test.new():
        yield


tag = Tag()


class A:
    pass


class B:
    pass


def test_tags():
    @factory(tags=[tag])
    def build_a() -> A:
        return A()

    class BFactory(Factory):
        __antidote__ = Factory.Conf(tags=[tag])

        def __call__(self) -> B:
            return B()

    class X(Service):
        __antidote__ = Service.Conf(tags=[tag])

    tagged = cast(Tagged, world.get(tag))

    assert len(tagged) == 3
    assert list(tagged.tags()) == [tag, tag, tag]
    assert set(tagged.values()) == {world.get(A @ build_a),
                                    world.get(B @ BFactory),
                                    world.get(X)}


def test_multiple_tags():
    tag2 = Tag()
    tags = [tag, tag2]

    @factory(tags=tags)
    def build_a() -> A:
        return A()

    class BFactory(Factory):
        __antidote__ = Factory.Conf(tags=tags)

        def __call__(self) -> B:
            return B()

    class X(Service):
        __antidote__ = Service.Conf(tags=tags)

    for t in tags:
        tagged = cast(Tagged, world.get(t))
        assert len(tagged) == 3
        assert list(tagged.tags()) == [t, t, t]
        assert set(tagged.values()) == {world.get(A @ build_a),
                                        world.get(B @ BFactory),
                                        world.get(X)}


def test_custom_tags():
    class CustomTag(Tag):
        __slots__ = ('name', 'attr')
        name: str
        attr: int

        def __init__(self, name: str, attr: int = 0):
            super().__init__(name=name, attr=attr)

        def group(self):
            return self.name

    tag1 = CustomTag('my_tag', attr=1)

    @factory(tags=[tag1])
    def build_a() -> A:
        return A()

    tag2 = CustomTag('my_tag', attr=2)

    class BFactory(Factory):
        __antidote__ = Factory.Conf(tags=[tag2])

        def __call__(self) -> B:
            return B()

    tag3 = CustomTag('my_tag', attr=3)

    class X(Service):
        __antidote__ = Service.Conf(tags=[tag3])

    tagged = cast(Tagged, world.get(CustomTag('my_tag')))
    assert len(tagged) == 3
    assert set(tagged.items()) == {
        (tag1, world.get(A @ build_a)),
        (tag2, world.get(B @ BFactory)),
        (tag3, world.get(X))
    }



from typing import Callable

import pytest

from antidote import world
from antidote._providers import IndirectProvider, ServiceProvider
from antidote.core.exceptions import DependencyNotFoundError
from antidote.exceptions import DuplicateDependencyError, FrozenWorldError


class Interface:
    pass


class A(Interface):
    pass


class B(Interface):
    pass


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        yield


@pytest.fixture
def service(empty_world):
    world.provider(ServiceProvider)
    return world.get(ServiceProvider)


@pytest.fixture
def indirect(empty_world):
    world.provider(IndirectProvider)
    return world.get(IndirectProvider)


def test_simple():
    indirect = IndirectProvider()
    world.singletons.add(A, A())
    indirect.register_static(Interface, A)

    assert world.test.maybe_provide_from(indirect, Interface).value is world.get(A)
    assert world.test.maybe_provide_from(indirect, Interface).singleton is True
    assert str(Interface) in repr(indirect)


def test_static_exists():
    indirect = IndirectProvider()
    world.singletons.add(A, A())
    indirect.register_static(Interface, A)

    assert not indirect.exists(object())
    assert indirect.exists(Interface)
    assert not indirect.exists(A)


def test_link_exists():
    indirect = IndirectProvider()
    world.singletons.add(A, A())
    indirect.register_link(Interface, lambda: A)

    assert not indirect.exists(object())
    assert indirect.exists(Interface)
    assert not indirect.exists(A)


def test_not_singleton_interface(service: ServiceProvider):
    indirect = IndirectProvider()
    service.register(A, singleton=False)
    indirect.register_static(Interface, A)

    assert world.test.maybe_provide_from(indirect, Interface).value is not world.get(A)
    assert isinstance(world.test.maybe_provide_from(indirect, Interface).value, A)
    assert world.test.maybe_provide_from(indirect, Interface).singleton is False


@pytest.mark.parametrize('singleton', [True, False])
@pytest.mark.parametrize('permanent', [True, False])
def test_link_permanent_singleton(service: ServiceProvider,
                                  singleton: bool,
                                  permanent: bool):
    choice = 'a'

    def implementation():
        return dict(a=A, b=B)[choice]

    service.register(A, singleton=singleton)
    service.register(B, singleton=singleton)
    indirect = IndirectProvider()
    indirect.register_link(Interface, linker=implementation, permanent=permanent)

    instance = world.test.maybe_provide_from(indirect, Interface).value
    assert isinstance(instance, A)
    assert (instance is world.get(A)) is singleton
    assert world.test.maybe_provide_from(indirect,
                                         Interface).singleton is (singleton and permanent)

    choice = 'b'
    assert implementation() == B
    assert world.test.maybe_provide_from(indirect,
                                         Interface).singleton is (singleton and permanent)
    instance = world.test.maybe_provide_from(indirect, Interface).value
    if permanent:
        assert isinstance(instance, A)
        assert (instance is world.get(A)) is singleton
    else:
        assert isinstance(instance, B)
        assert (instance is world.get(B)) is singleton


@pytest.mark.parametrize('keep_singletons_cache', [True, False])
@pytest.mark.parametrize('register', [
    pytest.param(lambda indirect, inf, impl: indirect.register_static(inf, impl),
                 id='register_static'),
    pytest.param(
        lambda indirect, inf, impl: indirect.register_link(inf, lambda: impl,
                                                           permanent=False),
        id='register_link'),
    pytest.param(
        lambda indirect, inf, impl: indirect.register_link(inf, lambda: impl,
                                                           permanent=True),
        id='register_link_permanent')
])
def test_copy(indirect: IndirectProvider,
              keep_singletons_cache: bool,
              register: Callable[[IndirectProvider, type, type], object]):
    world.singletons.add({A: A()})

    register(indirect, Interface, A)
    a = world.get(Interface)
    assert isinstance(a, Interface)

    if keep_singletons_cache:
        with world.test.clone(keep_singletons=True):
            clone = indirect.clone(True)
            assert world.test.maybe_provide_from(clone, Interface).value is a
    else:
        with world.test.empty():
            world.singletons.add({A: A(), B: B()})
            clone = indirect.clone(False)
            instance = world.test.maybe_provide_from(clone, Interface).value
            assert instance is world.get(A)
            assert instance is not a

    class Interface2:
        pass

    class A2(Interface2):
        pass

    class Interface3:
        pass

    class A3(Interface3):
        pass

    world.singletons.add({A2: A2(), A3: A3()})
    # Original does not modify clone
    register(indirect, Interface2, A2)
    assert world.get(Interface2) is world.get(A2)
    assert world.test.maybe_provide_from(clone, Interface2) is None

    # Did not modify original provider
    register(clone, Interface3, A3)
    assert world.test.maybe_provide_from(clone, Interface3).value is world.get(A3)
    with pytest.raises(DependencyNotFoundError):
        world.get(Interface3)


def test_freeze(indirect: IndirectProvider):
    world.freeze()

    with pytest.raises(FrozenWorldError):
        indirect.register_link(Interface, lambda: A, permanent=False)

    with pytest.raises(FrozenWorldError):
        indirect.register_static(Interface, A)

    with pytest.raises(DependencyNotFoundError):
        world.get(Interface)


def test_register_static_duplicate_check(indirect: IndirectProvider):
    indirect.register_static(Interface, A)

    with pytest.raises(DuplicateDependencyError):
        indirect.register_static(Interface, A)

    with pytest.raises(DuplicateDependencyError):
        indirect.register_link(Interface, lambda: A)


def test_register_duplicate_check(indirect: IndirectProvider):
    indirect.register_link(Interface, lambda: A)

    with pytest.raises(DuplicateDependencyError):
        indirect.register_static(Interface, A)

    with pytest.raises(DuplicateDependencyError):
        indirect.register_link(Interface, lambda: A)


def test_invalid_link():
    indirect = IndirectProvider()
    indirect.register_static('A', 'B')

    with pytest.raises(DependencyNotFoundError, match=".*B.*"):
        world.test.maybe_provide_from(indirect, 'A')

    for permanent in [True, False]:
        origin = f'origin-{permanent}'
        target = f'target-{permanent}'

        def impl():
            return target

        indirect.register_link(origin, linker=impl, permanent=permanent)

        with pytest.raises(DependencyNotFoundError, match=".*" + target + ".*"):
            world.test.maybe_provide_from(indirect, origin)



import functools
import inspect
from typing import Callable, Iterable, TypeVar, Union

from ._implementation import ImplementationMeta
from ._internal import API
from ._providers import IndirectProvider
from .core import inject
from .core.injection import DEPENDENCIES_TYPE
from .service import Service

F = TypeVar('F', bound=Callable[[], type])
C = TypeVar('C', bound=type)


@API.experimental
class Implementation(Service, metaclass=ImplementationMeta, abstract=True):
    """
    Essentially syntactic sugar to define easily a single implementation for an interface.
    The class will automatically be defined as a :py:class:`~.Service`, hence it has the
    same features.

    .. doctest:: helpers_implementation_class

        >>> from antidote import world, Implementation
        >>> class Interface:
        ...     pass
        >>> class Impl(Interface, Implementation):
        ...     pass
        >>> world.get(Interface)
        <Impl ...>

    .. note::

        If you need to chose between multiple implementations use
        :py:func:`.implementation`.
    """


@API.public
def implementation(interface: type,
                   *,
                   permanent: bool = True,
                   auto_wire: bool = True,
                   dependencies: DEPENDENCIES_TYPE = None,
                   use_names: Union[bool, Iterable[str]] = None,
                   use_type_hints: Union[bool, Iterable[str]] = None) -> Callable[[F], F]:
    """
    Function decorator which decides which implementation should be used for
    :code:`interface`.

    The underlying function is expected to return a dependency, typically the class of
    the implementation when defined as a service. You may also use a factory dependency.

    The function will not be wired, you'll need to do it yourself if you need it.

    .. doctest:: helpers_implementation

        >>> from antidote import implementation, Service, factory, world
        >>> class Interface:
        ...     pass
        >>> class A(Interface, Service):
        ...     pass
        >>> class B(Interface):
        ...     pass
        >>> @factory
        ... def build_b() -> B:
        ...     return B()
        >>> @implementation(Interface, dependencies=['choice'])
        ... def choose_interface(choice: str):
        ...     if choice == 'a':
        ...         return A  # One could also use A.with_kwargs(...)
        ...     else:
        ...         return B @ build_b  # or B @ build_b.with_kwargs(...)
        >>> world.singletons.add('choice', 'b')
        >>> world.get(Interface)
        <B ...>
        >>> # Changing choice doesn't matter anymore as the implementation is permanent.
        ... with world.test.clone(overridable=True):
        ...     world.singletons.add('choice', 'a')
        ...     world.get(Interface)
        <B ...>

    Args:
        interface: Interface for which an implementation will be provided
        permanent: Whether the function should be called each time the interface is needed
            or not. Defaults to :py:obj:`True`.
        auto_wire: Whether the function should have its arguments injected or not
            with :py:func:`~.injection.inject`.
        dependencies: Propagated to :py:func:`~.injection.inject`.
        use_names: Propagated to :py:func:`~.injection.inject`.
        use_type_hints: Propagated to :py:func:`~.injection.inject`.

    Returns:
        The decorated function, unmodified.
    """
    if not inspect.isclass(interface):
        raise TypeError(f"interface must be a class, not a {type(interface)}")

    if not (auto_wire is None or isinstance(auto_wire, bool)):
        raise TypeError(f"auto_wire can be None or a boolean, not {type(auto_wire)}")

    @inject
    def register(func, indirect_provider: IndirectProvider):
        from ._providers.factory import FactoryDependency
        from ._providers.service import Build

        if inspect.isfunction(func):
            if auto_wire:
                func = inject(func,
                              dependencies=dependencies,
                              use_names=use_names,
                              use_type_hints=use_type_hints)

            @functools.wraps(func)
            def linker():
                dependency = func()
                cls = dependency
                if isinstance(cls, Build):
                    cls = cls.dependency
                if isinstance(cls, FactoryDependency):
                    cls = cls.dependency

                if not (isinstance(cls, type) and inspect.isclass(cls)
                        and issubclass(cls, interface)):
                    raise TypeError(f"{func} is expected to return a class or a "
                                    f"Service / Factory dependency which implements"
                                    f"{interface}")
                return dependency

            indirect_provider.register_link(interface, linker=linker, permanent=permanent)
        else:
            raise TypeError(f"implementation must be applied on a function, "
                            f"not a {type(func)}")
        return func

    return register



# @formatter:off
from cpython.ref cimport PyObject

from antidote._internal.stack cimport DependencyStack
# @formatter:on

# flags
cdef:
    size_t FLAG_DEFINED, FLAG_SINGLETON

cdef class PyObjectBox:
    cdef:
        object obj

cdef struct DependencyResult:
    PyObject*box
    size_t flags

cdef struct ProviderCache:
    size_t length
    size_t capacity
    PyObject** dependencies
    size_t*counters
    PyObject** providers

cdef class DependencyInstance:
    cdef:
        readonly bint singleton
        readonly object value

cdef class Container:
    pass

cdef class RawContainer(Container):
    cdef:
        list _providers
        dict _singletons
        object _singleton_lock
        unsigned long _singletons_clock
        DependencyStack _dependency_stack
        ProviderCache __cache
        bint __frozen
        object __freeze_lock
        object __weakref__

    cdef fast_get(self, PyObject*dependency, DependencyResult*result)
    cdef __safe_provide(self, PyObject*dependency, DependencyResult*result,
                        unsigned long singletons_clock)

cdef class RawProvider:
    cdef:
        object _container_ref

    cdef fast_provide(self, PyObject*dependency, PyObject*container,
                      DependencyResult*result)

cdef class FastProvider(RawProvider):
    pass



import pytest

from antidote import world
from antidote._providers import FactoryProvider
from antidote._providers.service import Build
from antidote.exceptions import (DependencyNotFoundError, DuplicateDependencyError,
                                 FrozenWorldError)


@pytest.fixture()
def provider():
    with world.test.empty():
        world.provider(FactoryProvider)
        yield world.get(FactoryProvider)


class A:
    def __init__(self, **kwargs):
        self.kwargs = kwargs


def build(**kwargs) -> A:
    return A(**kwargs)


def test_str(provider: FactoryProvider):
    provider.register(A, build)
    assert str(A) in str(provider)
    assert str(build) in str(provider)


def test_simple(provider: FactoryProvider):
    factory_id = provider.register(A, build)
    assert isinstance(world.get(factory_id), A)
    # singleton
    assert world.get(factory_id) is world.get(factory_id)


def test_lazy(provider: FactoryProvider):
    world.singletons.add('A', build)
    factory_id = provider.register(A, world.lazy('A'))
    assert isinstance(world.get(factory_id), A)
    # singleton
    assert world.get(factory_id) is world.get(factory_id)


def test_exists():
    provider = FactoryProvider()
    factory_id = provider.register(A, build)

    assert not provider.exists(object())
    assert not provider.exists(Build(object(), kwargs=dict(a=1)))
    assert provider.exists(factory_id)
    assert provider.exists(Build(factory_id, kwargs=dict(a=1)))
    assert not provider.exists(A)
    assert not provider.exists(build)


@pytest.mark.parametrize('singleton', [True, False])
def test_singleton(singleton: bool):
    with world.test.empty():
        world.singletons.add('A', build)
        world.provider(FactoryProvider)
        with world.test.clone():
            provider = world.get(FactoryProvider)
            factory_id = provider.register(A, build, singleton=singleton)
            assert isinstance(world.get(factory_id), A)
            assert (world.get(factory_id) is world.get(factory_id)) == singleton

        with world.test.clone(keep_singletons=True):
            provider = world.get(FactoryProvider)
            factory_id = provider.register(A, world.lazy('A'), singleton=singleton)
            assert isinstance(world.get(factory_id), A)
            assert (world.get(factory_id) is world.get(factory_id)) == singleton


@pytest.mark.parametrize('first', ['register', 'register_lazy'])
@pytest.mark.parametrize('second', ['register', 'register_lazy'])
def test_duplicate_dependency(provider: FactoryProvider, first: str, second: str):
    world.singletons.add('A', build)

    if first == 'register':
        provider.register(A, build)
    else:
        provider.register(A, world.lazy('A'))

    with pytest.raises(DuplicateDependencyError):
        if second == 'register':
            provider.register(A, build)
        else:
            provider.register(A, world.lazy('A'))


def test_invalid_register(provider: FactoryProvider):
    with pytest.raises(TypeError):
        provider.register(A, 1)


def test_build_dependency(provider: FactoryProvider):
    world.singletons.add('A', build)
    kwargs = dict(test=object())

    with world.test.clone():
        provider = world.get(FactoryProvider)
        factory_id = provider.register(A, build)
        a = world.get(Build(factory_id, kwargs))
        assert isinstance(a, A)
        assert a.kwargs == kwargs

    with world.test.clone(keep_singletons=True):
        provider = world.get(FactoryProvider)
        factory_id = provider.register(A, world.lazy('A'))
        a = world.get(Build(factory_id, kwargs))
        assert isinstance(a, A)
        assert a.kwargs == kwargs


def test_copy():
    class B:
        pass

    class C:
        pass

    def build2() -> A:
        return A(build2=True)

    with world.test.empty():
        original = FactoryProvider()
        a_id = original.register(A, build)

        # cloned has the same dependencies
        cloned = original.clone(False)
        with pytest.raises(DuplicateDependencyError):
            cloned.register(A, build)
        assert isinstance(world.test.maybe_provide_from(cloned, a_id).value, A)

        # Adding dependencies to either original or cloned, should not impact the
        # other one.
        _ = original.register(B, lambda: B())
        _ = cloned.register(B, lambda: B())

        cloned.register(C, lambda: C())
        original.register(C, lambda: C())

    with world.test.empty():
        world.singletons.add('build', build)
        original = FactoryProvider()
        a_id = original.register(A, world.lazy('build'))
        a = world.test.maybe_provide_from(original, a_id).value

        assert isinstance(a, A)
        assert a.kwargs == {}

        cloned_with_singletons = original.clone(True)
        cloned = original.clone(False)
        with world.test.empty():
            world.singletons.add('build', build2)
            a2 = world.test.maybe_provide_from(cloned, a_id).value

            assert isinstance(a2, A)
            assert a2.kwargs == dict(build2=True)

            # We kept singletons, so previous dependency 'build' has been kept.
            a3 = world.test.maybe_provide_from(cloned_with_singletons, a_id).value
            assert isinstance(a3, A)
            assert a3.kwargs == {}


def test_freeze(provider: FactoryProvider):
    world.freeze()

    with pytest.raises(FrozenWorldError):
        provider.register(A, build)


def test_factory_id_repr(provider: FactoryProvider):
    factory_id = provider.register(A, build)
    assert f"{__name__}.A" in repr(factory_id)
    assert f"{__name__}.build" in repr(factory_id)

    class B:
        pass

    factory_id = provider.register(B, world.lazy(build))
    assert f"{__name__}.test_factory_id_repr.<locals>.B" in repr(factory_id)
    assert f"{__name__}.build" in repr(factory_id)


def test_unknown_dependency():
    p1 = FactoryProvider()
    assert world.test.maybe_provide_from(p1, object()) is None
    assert p1.maybe_debug(object()) is None

    p2 = FactoryProvider()
    dependency = p2.register(A, build)
    assert world.test.maybe_provide_from(p1, dependency) is None
    assert p1.maybe_debug(dependency) is None


def test_invalid_lazy_dependency():
    provider = FactoryProvider()
    fid = provider.register('A', factory=world.lazy("factory"))

    with pytest.raises(DependencyNotFoundError, match=".*factory.*"):
        world.test.maybe_provide_from(provider, fid)



import base64
import inspect
import textwrap
from collections import deque
from typing import Deque, Hashable, List, Sequence, Tuple, TYPE_CHECKING

from .immutable import Immutable
from .. import API

if TYPE_CHECKING:
    from ...core.container import RawContainer

# Object will be allocated on the heap, so as close as possible to most user objects
# in memory.
_ID_MASK = id(object())


@API.private
def short_id(obj) -> str:
    """ Produces a short, human readable, representation of the id of an object. """
    n = id(obj) ^ _ID_MASK
    return (base64
            .b64encode(n.to_bytes(8, byteorder='little'))
            .decode('ascii')
            .rstrip('=')  # Remove padding
            .rstrip('A'))  # Remove 000000


@API.private
def debug_repr(x) -> str:
    from ..wrapper import is_wrapper
    try:
        return x.__antidote_debug_repr__()
    except AttributeError:
        pass
    if (isinstance(x, type) and inspect.isclass(x)) \
            or inspect.isfunction(x) \
            or is_wrapper(x):
        module = (x.__module__ + ".") if x.__module__ != "__main__" else ""
        return f"{module}{x.__qualname__}"
    return repr(x)


@API.private
def get_injections(func) -> Sequence:
    from ..wrapper import get_wrapper_dependencies
    try:
        return get_wrapper_dependencies(func)
    except TypeError:
        return []


@API.private
class DebugTreeNode(Immutable):
    __slots__ = ('info', 'singleton', 'children')
    info: str
    singleton: bool
    children: 'List[DebugTreeNode]'

    def __init__(self, info: str, *, singleton: bool = True,
                 children: 'List[DebugTreeNode]' = None):
        super().__init__(info=textwrap.dedent(info),
                         singleton=singleton,
                         children=children or [])


@API.private
class Task(Immutable):
    __slots__ = ()


@API.private
class DependencyTask(Task):
    __slots__ = ('dependency',)
    dependency: object


@API.private
class InjectionTask(Task):
    __slots__ = ('name', 'injections')
    name: str
    injections: List[Hashable]


@API.private
def tree_debug_info(container: 'RawContainer',
                    origin,
                    depth: int = -1) -> str:
    from ...core.wiring import WithWiringMixin
    from ...core.exceptions import DependencyNotFoundError

    if depth < 0:
        depth = 1 << 31  # roughly infinity in this case.

    depth += 1  # To match root = depth 0
    root = DebugTreeNode(info=debug_repr(origin))
    original_root = root
    tasks: Deque[Tuple[DebugTreeNode, set, Task]] = deque([
        (root, set(), DependencyTask(origin))
    ])

    def add_root_injections(parent: DebugTreeNode, parent_dependencies: set, dependency):
        if isinstance(dependency, type) and inspect.isclass(dependency):
            cls = dependency
            conf = getattr(cls, '__antidote__', None)
            if conf is not None \
                    and isinstance(conf, WithWiringMixin) \
                    and conf.wiring is not None:
                for m in sorted(conf.wiring.methods):
                    if m != '__init__':
                        tasks.append((parent, parent_dependencies, InjectionTask(
                            name=f"Method: {m}",
                            injections=get_injections(getattr(cls, m)),
                        )))
        elif callable(dependency):
            for d in get_injections(dependency):
                tasks.append((parent, parent_dependencies, DependencyTask(d)))

    while tasks:
        parent, parent_dependencies, task = tasks.pop()
        if isinstance(task, DependencyTask):
            dependency = task.dependency
            try:
                debug = container.debug(dependency)
            except DependencyNotFoundError:
                if dependency is origin:
                    add_root_injections(parent, parent_dependencies, dependency)
                else:
                    parent.children.append(DebugTreeNode(f"/!\\ Unknown: "
                                                         f"{debug_repr(dependency)}"))

                continue

            if dependency in parent_dependencies:
                parent.children.append(DebugTreeNode(f"/!\\ Cyclic dependency: "
                                                     f"{debug.info}"))
                continue

            tree_node = DebugTreeNode(debug.info, singleton=debug.singleton)

            parent.children.append(tree_node)
            parent = tree_node
            parent_dependencies = parent_dependencies | {dependency}

            if dependency is origin:
                root = tree_node  # previous root is redundant
                add_root_injections(parent, parent_dependencies, dependency)

            if len(parent_dependencies) < depth:
                for d in debug.dependencies:
                    tasks.append((parent, parent_dependencies,
                                  DependencyTask(d)))
                for w in debug.wired:
                    if isinstance(w, type) and inspect.isclass(w):
                        for d in get_injections(getattr(w, '__init__')):
                            tasks.append((parent, parent_dependencies,
                                          DependencyTask(d)))
                    else:
                        tasks.append((parent, parent_dependencies, InjectionTask(
                            name=debug_repr(w),
                            injections=get_injections(w),
                        )))
        elif isinstance(task, InjectionTask) and task.injections:
            tree_node = DebugTreeNode(task.name)
            parent.children.append(tree_node)
            parent = tree_node
            for d in task.injections:
                tasks.append((parent, parent_dependencies, DependencyTask(d)))

    if not root.children and original_root is root:
        return f"{origin!r} is neither a dependency nor is anything injected."

    output = [
        ("* " if not root.singleton else "") + root.info
    ]
    nodes: Deque[Tuple[str, bool, DebugTreeNode]] = deque([
        ("", i == 0, child)
        for i, child in enumerate(root.children[::-1])
    ])
    any_not_singleton = not root.singleton

    while nodes:
        prefix, last, node = nodes.pop()
        any_not_singleton |= (not node.singleton)
        first_line, *rest = node.info.split("\n", 1)
        txt = prefix + ("└──" if last else "├──")
        txt += (" * " if not node.singleton else " ") + first_line
        new_prefix = prefix + ("    " if last else "│   ")
        if rest:
            txt += "\n" + textwrap.indent(rest[0], new_prefix)
        output.append(txt)

        for i, child in enumerate(node.children[::-1]):
            nodes.append((new_prefix, i == 0, child))

    output.append("")
    if any_not_singleton:
        output.extend(["* = not singleton",
                       ""])

    return "\n".join(output)



import itertools
from typing import Callable

import pytest

from antidote import factory, implementation, world
from antidote._compatibility.typing import Protocol


class FactoryOutput:
    def __init__(self, a, b):
        self.a = a
        self.b = b


class Interface:
    pass


class A:
    pass


class B:
    pass


class FuncProtocol(Protocol):
    def __call__(self, a: A, b: B = None) -> FactoryOutput:
        pass


@pytest.fixture(autouse=True)
def new_world():
    with world.test.new():
        world.singletons.add({A: A(),
                              B: B(),
                              'a': object(),
                              'b': object(),
                              'x': object()})
        yield


class WiringTestCase:
    def __init__(self, func: FuncProtocol, with_type_hints: bool):
        self.func = func
        self.with_type_hints = with_type_hints


def impl_interface(**kwargs):
    return implementation(Interface, **kwargs)


@pytest.fixture(params=[
    pytest.param((f, tp), id=f"{f.__name__} - {tp}")
    for (f, tp) in itertools.product([factory, impl_interface],
                                     ['type_hints', None])
])
def builder(request):
    (wrapper, with_type_hints) = request.param

    if with_type_hints:
        def func(a: A, b: B = None) -> FactoryOutput:
            return FactoryOutput(a, b)
    else:
        def func(a, b=None) -> FactoryOutput:
            return FactoryOutput(a, b)

    def build(**kwargs):
        return WiringTestCase(wrapper(**kwargs)(func),
                              with_type_hints=with_type_hints is not None)

    return build


Builder = Callable[..., WiringTestCase]


def test_auto_wiring(builder: Builder):
    test_case = builder()
    if test_case.with_type_hints:
        out = test_case.func()
        assert out.a is world.get(A)
        assert out.b is world.get(B)
    else:
        with pytest.raises(TypeError):
            test_case.func()


def test_no_auto_wiring(builder: Builder):
    test_case = builder(auto_wire=False)

    with pytest.raises(TypeError):
        test_case.func()

    a = object()
    out = test_case.func(a)
    assert out.a is a
    assert out.b is None


def test_use_names(builder: Builder):
    with world.test.clone(keep_singletons=True):
        test_case = builder(use_names=False)

        if test_case.with_type_hints:
            out = test_case.func()
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            with pytest.raises(TypeError):
                test_case.func()

    with world.test.clone(keep_singletons=True):
        test_case = builder(use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('a')
            assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(use_names=['a'])

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('a')
            assert out.b is None


def test_use_type_hints(builder: Builder):
    with world.test.clone(keep_singletons=True):
        test_case = builder(use_type_hints=True)

        if test_case.with_type_hints:
            out = test_case.func()
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            with pytest.raises(TypeError):
                test_case.func()

    with world.test.clone(keep_singletons=True):
        test_case = builder(use_type_hints=False)
        with pytest.raises(TypeError):
            test_case.func()

    with world.test.clone(keep_singletons=True):
        test_case = builder(use_type_hints=False, use_names=True)
        out = test_case.func()
        assert out.a is world.get('a')
        assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(use_type_hints=['a'])

        if test_case.with_type_hints:
            out = test_case.func()
            assert out.a is world.get(A)
            assert out.b is None
        else:
            with pytest.raises(TypeError):
                test_case.func()


def test_dependencies_dict(builder: Builder):
    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=dict(), use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('a')
            assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=dict(a='x'), use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get('x')
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('x')
            assert out.b is world.get('b')


def test_dependencies_seq(builder: Builder):
    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=[], use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('a')
            assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=['x'], use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get('x')
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('x')
            assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=['x', None], use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get('x')
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('x')
            assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=[None, 'x'], use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            out = test_case.func()
            assert out.a is world.get(A)
            assert out.b is world.get('x')
        else:
            out = test_case.func()
            assert out.a is world.get('a')
            assert out.b is world.get('x')


def test_dependencies_callable(builder: Builder):
    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=lambda arg: None, use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get(A)
            assert out.b is world.get(B)
        else:
            assert out.a is world.get('a')
            assert out.b is world.get('b')

    with world.test.clone(keep_singletons=True):
        test_case = builder(dependencies=lambda arg: 'x' if arg.name == 'b' else None,
                            use_names=True)

        out = test_case.func()
        if test_case.with_type_hints:
            assert out.a is world.get(A)
            assert out.b is world.get('x')
        else:
            assert out.a is world.get('a')
            assert out.b is world.get('x')


def test_dependencies_str(builder: Builder):
    with world.test.clone(keep_singletons=True):
        world.singletons.add({
            'conf:a': object(),
            'conf:b': object()
        })
        out = builder(dependencies='conf:{arg_name}').func()
        assert out.a is world.get('conf:a')
        assert out.b is world.get('conf:b')


@pytest.mark.parametrize('func', [factory, impl_interface])
@pytest.mark.parametrize('expectation,kwargs', [
    pytest.param(pytest.raises(TypeError, match=f".*{arg}.*"),
                 {arg: object()},
                 id=arg)
    for arg in ['auto_wire', 'dependencies', 'use_names', 'use_type_hints']
])
def test_invalid(func, expectation, kwargs):
    def f() -> FactoryOutput:
        pass

    with expectation:
        func(**kwargs)(f)



import pytest

from antidote import world
from antidote._providers import ServiceProvider
from antidote.core import (Dependency)
from antidote.core.exceptions import DuplicateDependencyError
from antidote.exceptions import DependencyNotFoundError, FrozenWorldError
from .utils import DummyIntProvider


@pytest.fixture(autouse=True)
def empty_world():
    with world.test.empty():
        yield


class A:
    pass


def test_singletons():
    world.singletons.add("singleton", 12342)
    assert world.get("singleton") == 12342

    world.singletons.add({
        "singleton2": 89,
        "singleton3": 123
    })
    assert world.get("singleton2") == 89
    assert world.get("singleton3") == 123


def test_invalid_singletons():
    with pytest.raises(TypeError):
        world.singletons.add(1)


def test_duplicate_singletons():
    world.singletons.add("singleton", 12342)

    with pytest.raises(DuplicateDependencyError, match=".*singleton.*12342.*"):
        world.singletons.add("singleton", 1)

    with pytest.raises(DuplicateDependencyError, match=".*singleton.*12342.*"):
        world.singletons.add({"singleton": 1})


def test_get():
    world.singletons.add("x", 1)
    assert 1 == world.get("x")

    with pytest.raises(DependencyNotFoundError):
        world.get("nothing")

    world.singletons.add(A, A())
    assert world.get[int]("x") == 1
    assert world.get[A]() is world.get(A)


def test_lazy():
    world.singletons.add({
        'x': object(),
        A: A()
    })

    lazy = world.lazy('x')
    assert isinstance(lazy, Dependency)
    assert lazy.value == 'x'
    assert lazy.get() == world.get('x')

    lazy = world.lazy[int]('x')
    assert isinstance(lazy, Dependency)
    assert lazy.value == 'x'
    assert lazy.get() == world.get('x')
    assert world.lazy[A]().get() is world.get(A)


def test_freeze():
    world.provider(ServiceProvider)
    factory = world.get(ServiceProvider)

    class Service:
        pass

    world.freeze()
    with pytest.raises(FrozenWorldError):
        world.singletons.add("test", "x")

    with pytest.raises(FrozenWorldError):
        factory.register(Service)


def test_add_provider():
    world.provider(DummyIntProvider)
    assert world.get(10) == 20


@pytest.mark.parametrize('p, expectation', [
    (1, pytest.raises(TypeError)),
    (A, pytest.raises(TypeError))
])
def test_invalid_add_provider(p, expectation):
    with expectation:
        world.provider(p)



from typing import Callable, Dict, Hashable, Optional

# @formatter:off
cimport cython
from cpython.object cimport PyObject, PyObject_CallObject

from antidote.core.container cimport (DependencyResult, FastProvider,
                                      FLAG_SINGLETON, RawContainer)
from .._internal.utils import debug_repr
from ..core.exceptions import DependencyNotFoundError
from ..core.utils import DependencyDebug

# @formatter:on

cdef extern from "Python.h":
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1
    PyObject*PyDict_GetItem(PyObject *p, PyObject *key)


@cython.final
cdef class IndirectProvider(FastProvider):
    cdef:
        dict __static_links
        dict __links

    def __init__(self):
        super().__init__()
        self.__links = dict()  # type: Dict[Hashable, Link]
        self.__static_links = dict()  # type: Dict[Hashable, Hashable]

    def __repr__(self):
        return f"{type(self).__name__}(links={self.__links}, " \
               f"static_links={self.__static_links})"

    def clone(self, keep_singletons_cache: bool) -> IndirectProvider:
        p = IndirectProvider()
        p.__links = self.__links.copy()
        p.__static_links = self.__static_links.copy()
        return p

    def exists(self, dependency) -> bool:
        return dependency in self.__static_links or dependency in self.__links

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        cdef:
            Link link

        try:
            link = self.__links[dependency]
        except KeyError:
            pass
        else:
            repr_d = debug_repr(dependency)
            repr_linker = debug_repr(link.linker)
            if link.permanent:
                if dependency in self.__static_links:
                    target = self.__static_links[dependency]
                    return DependencyDebug(
                        f"Permanent link: {repr_d} -> {debug_repr(target)} "
                        f"defined by {repr_linker}",
                        singleton=True,
                        dependencies=[target])
                else:
                    return DependencyDebug(
                        f"Permanent link: {repr_d} -> ??? "
                        f"defined by {repr_linker}",
                        singleton=True)
            else:
                return DependencyDebug(
                    f"Dynamic link: {repr_d} -> ??? defined by {repr_linker}",
                    singleton=False,
                    wired=[link.linker])

        try:
            target = self.__static_links[dependency]
        except KeyError:
            pass
        else:
            repr_d = debug_repr(dependency)
            return DependencyDebug(f"Static link: {repr_d} -> {debug_repr(target)}",
                                   singleton=True,
                                   dependencies=[target])

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        cdef:
            PyObject*ptr
            object target

        ptr = PyDict_GetItem(<PyObject*> self.__static_links, dependency)
        if ptr != NULL:
            (<RawContainer> container).fast_get(ptr, result)
            if result.flags == 0:
                raise DependencyNotFoundError(<object> ptr)
        else:
            ptr = PyDict_GetItem(<PyObject*> self.__links, dependency)
            if ptr != NULL:
                target = PyObject_CallObject((<Link> ptr).linker, <object> NULL)
                (<RawContainer> container).fast_get(<PyObject*> target, result)
                if result.flags == 0:
                    raise DependencyNotFoundError(target)
                result.flags &= (<Link> ptr).singleton_flag
                if (<Link> ptr).permanent:
                    PyDict_SetItem(<PyObject*> self.__static_links,
                                   dependency,
                                   <PyObject*> target)

    def register_static(self, dependency: Hashable, target_dependency: Hashable):
        with self._ensure_not_frozen():
            self._raise_if_exists(dependency)
            self.__static_links[dependency] = target_dependency

    def register_link(self, dependency: Hashable, linker: Callable[[], Hashable],
                      permanent: bool = True):
        with self._ensure_not_frozen():
            self._raise_if_exists(dependency)
            self.__links[dependency] = Link(linker, permanent)

@cython.final
cdef class Link:
    cdef:
        object linker
        bint permanent
        int singleton_flag

    def __init__(self, linker: Callable[[], Hashable], permanent: bool):
        self.linker = linker
        self.permanent = permanent
        self.singleton_flag = ~0 if permanent else ~FLAG_SINGLETON



import inspect
from typing import Dict, Hashable

# @formatter:off
cimport cython
from cpython.object cimport PyObject_Call, PyObject_CallObject
from cpython.ref cimport PyObject

from antidote.core.container cimport (DependencyResult, FastProvider,
                                      FLAG_DEFINED, FLAG_SINGLETON, PyObjectBox )
from .._internal.utils import debug_repr
# @formatter:on
from ..core.utils import DependencyDebug

cdef extern from "Python.h":
    int PyObject_IsInstance(PyObject *inst, PyObject *cls) except -1
    PyObject*PyDict_GetItem(PyObject *p, PyObject *key)
    PyObject*Py_True


@cython.final
cdef class Build:
    def __init__(self, dependency: Hashable, kwargs: Dict):
        assert isinstance(kwargs, dict) and len(kwargs) > 0
        self.dependency = dependency  # type: Hashable
        self.kwargs = kwargs  # type: Dict

        try:
            # Try most precise hash first
            self._hash = hash((self.dependency, tuple(self.kwargs.items())))
        except TypeError:
            # If type error, return the best error-free hash possible
            self._hash = hash((self.dependency, tuple(self.kwargs.keys())))

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return f"Build(dependency={self.dependency}, kwargs={self.kwargs})"

    def __antidote_debug_repr__(self):
        return f"{debug_repr(self.dependency)}(**{self.kwargs})"

    def __eq__(self, other):
        return (isinstance(other, Build)
                and self._hash == other._hash
                and (self.dependency is other.dependency
                     or self.dependency == other.dependency)
                and self.kwargs == other.kwargs)  # noqa

@cython.final
cdef class ServiceProvider(FastProvider):
    """
    Provider managing factories. Also used to register classes directly.
    """
    cdef:
        dict __services
        tuple __empty_tuple

    def __init__(self):
        super().__init__()
        self.__empty_tuple = tuple()
        self.__services = dict()  # type: Dict[Hashable, bool]

    def __repr__(self):
        return f"{type(self).__name__}(services={list(self.__services.items())!r})"

    def exists(self, dependency: Hashable) -> bool:
        if isinstance(dependency, Build):
            return dependency.dependency in self.__services
        return dependency in self.__services

    def clone(self, keep_singletons_cache: bool) -> ServiceProvider:
        p = ServiceProvider()
        p.__services = self.__services.copy()
        return p

    def maybe_debug(self, build: Hashable):
        klass = build.dependency if isinstance(build, Build) else build
        try:
            singleton = self.__services[klass]
        except KeyError:
            return None
        return DependencyDebug(debug_repr(build),
                               singleton=singleton,
                               wired=[klass])

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        cdef:
            PyObject*service
            PyObject*singleton
            tuple args
            object factory

        if PyObject_IsInstance(dependency, <PyObject*> Build):
            singleton = PyDict_GetItem(<PyObject*> self.__services,
                                       <PyObject*> (<Build> dependency).dependency)

            if singleton == NULL:
                return
            (<PyObjectBox> result.box).obj = PyObject_Call(
                (<Build> dependency).dependency,
                self.__empty_tuple,
                (<Build> dependency).kwargs
            )
        else:
            singleton = PyDict_GetItem(<PyObject*> self.__services, dependency)
            if singleton == NULL:
                return
            (<PyObjectBox> result.box).obj = PyObject_CallObject(<object> dependency,
                                                                 <object> NULL)

        result.flags = FLAG_DEFINED | (FLAG_SINGLETON if singleton == Py_True else 0)

    def register(self, klass: type, *, singleton: bool = True):
        if not (isinstance(klass, type) and inspect.isclass(klass)):
            raise TypeError(f"service must be a class, not {klass!r}")
        if not isinstance(singleton, bool):
            raise TypeError(f"singleton must be a boolean, not {singleton!r}")
        with self._ensure_not_frozen():
            self._raise_if_exists(klass)
            self.__services[klass] = singleton



from typing import Dict, Hashable

from antidote.core import DependencyInstance, Provider


class DummyProvider(Provider):
    singleton = True

    def clone(self, keep_singletons_cache: bool) -> Provider:
        return DummyProvider(self.data)

    def __init__(self, data: Dict = None):
        super().__init__()
        self.data = data

    def exists(self, dependency: Hashable) -> bool:
        return dependency in self.data

    def provide(self, dependency, container):
        return DependencyInstance(self.data[dependency],
                                  singleton=self.singleton)


class DummyFactoryProvider(Provider):
    singleton = True

    def clone(self, keep_singletons_cache: bool) -> Provider:
        return DummyFactoryProvider(self.data)

    def __init__(self, data: Dict = None):
        super().__init__()
        self.data = data or dict()

    def exists(self, dependency: Hashable) -> bool:
        return dependency in self.data

    def provide(self, dependency, container):
        return DependencyInstance(self.data[dependency](container),
                                  singleton=self.singleton)



from .. import API

_ABSTRACT_FLAG = '__antidote_abstract'


@API.private
class FinalMeta(type):
    def __new__(mcls, name, bases, namespace, **kwargs):
        for b in bases:
            if isinstance(b, FinalMeta):
                raise TypeError(f"Type '{b.__name__}' cannot be inherited.")
        return super().__new__(mcls, name, bases, namespace)


@API.private
class AbstractMeta(type):
    def __new__(mcls, name, bases, namespace, **kwargs):
        abstract = kwargs.get('abstract')
        namespace[_ABSTRACT_FLAG] = abstract
        if not abstract:
            for b in bases:
                if isinstance(b, mcls) and not getattr(b, _ABSTRACT_FLAG):
                    raise TypeError(
                        f"Cannot inherit a service which is not defined abstract. "
                        f"Consider defining {b} abstract by adding 'abstract=True' as a "
                        f"metaclass parameter. If so, Antidote won't use it.")

        return super().__new__(mcls, name, bases, namespace)



import pytest

from antidote import world
from antidote.core.wiring import Wiring, WithWiringMixin


class Dummy:
    pass


@pytest.fixture(autouse=True)
def new_world():
    with world.test.empty():
        world.singletons.add({'x': object(),
                              'xx': object(),
                              'y': object(),
                              'z': object(),
                              Dummy: Dummy()})
        yield


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(methods=[]), pytest.raises(ValueError, match=".*method.*")),
    (dict(methods=[1]), pytest.raises(ValueError, match="(?i).*method.*")),
    (dict(), pytest.raises(TypeError, match=".*.*")),
    (dict(methods=1), pytest.raises(TypeError, match=".*method.*")),
    (dict(methods=['f'], dependencies=1),
     pytest.raises(TypeError, match=".*dependencies.*")),
    (dict(methods=['f'], use_names=1), pytest.raises(TypeError, match=".*use_names.*")),
    (dict(methods=['f'], use_type_hints=1),
     pytest.raises(TypeError, match=".*use_type_hints.*")),
    (dict(methods=['f'], wire_super=1), pytest.raises(TypeError, match=".*wire_super.*")),
    (dict(methods=['f'], wire_super=['x']),
     pytest.raises(ValueError, match=".*wire_super.*")),
    (dict(methods=['f'], ignore_missing_method=1),
     pytest.raises(TypeError, match=".*ignore_missing_method.*")),
    (dict(methods=['f'], ignore_missing_method=['x']),
     pytest.raises(ValueError, match=".*ignore_missing_method.*"))
])
def test_validation(kwargs, expectation):
    with expectation:
        Wiring(**kwargs)


@pytest.mark.parametrize('kwargs, expected', [
    (dict(methods=['f']), dict()),
    (dict(methods=iter(['f', 'x'])), dict(methods={'f', 'x'})),
    (dict(methods=['f'], use_names=iter(['x'])), dict(use_names={'x'})),
    (dict(methods=['f'], use_type_hints=iter(['x'])), dict(use_type_hints={'x'})),
    (dict(methods=['f'], wire_super=iter(['f'])), dict(wire_super={'f'})),
    (dict(methods=['f'], wire_super=True), dict(wire_super={'f'})),
    (dict(methods=['f'], wire_super=False), dict(wire_super=set())),
    (dict(methods=['f'], ignore_missing_method=True),
     dict(ignore_missing_method={'f'})),
    (dict(methods=['f'], ignore_missing_method=False),
     dict(ignore_missing_method=set())),
    (dict(methods=['f'], ignore_missing_method=iter(['f'])),
     dict(ignore_missing_method={'f'})),
])
def test_init(kwargs, expected):
    result = Wiring(**kwargs)
    expected_methods = expected.get('methods', {'f'})
    assert result.methods == expected_methods
    assert result.wire_super == expected.get('wire_super', set())
    assert result.ignore_missing_method == expected.get('ignore_missing_method', set())
    assert result.dependencies == expected.get('dependencies')
    assert result.use_names == expected.get('use_names')
    assert result.use_type_hints == expected.get('use_type_hints')


def test_wiring_methods():
    wiring = Wiring(methods=['f'])

    @wiring.wire
    class A:
        def f(self, x: Dummy):
            return x

        def g(self, x: Dummy):
            return x

    assert A().f() is world.get(Dummy)
    with pytest.raises(TypeError):
        A().g()


def test_wiring_super_methods():
    class A:
        def f(self, x: Dummy):
            return x

    with pytest.raises(AttributeError, match=".*'f'.*"):
        wiring = Wiring(methods=['f'])

        @wiring.wire
        class X(A):
            pass

    with pytest.raises(AttributeError, match=".*'f'.*"):
        wiring = Wiring(methods=['f'], wire_super=False)

        @wiring.wire
        class Y(A):
            pass

    c_wiring = Wiring(methods=['f'], wire_super=True)

    @c_wiring.wire
    class C(A):
        pass

    assert C().f() is world.get(Dummy)


def test_wiring_dependencies():
    a_wiring = Wiring(methods=['f'], dependencies=('y',))

    @a_wiring.wire
    class A:
        def f(self, x: Dummy):
            return x

    assert A().f() is world.get('y')

    b_wiring = Wiring(methods=['f'], dependencies=dict(x='z'))

    @b_wiring.wire
    class B:
        def f(self, x: Dummy):
            return x

    assert B().f() is world.get('z')

    c_wiring = Wiring(methods=['f'], dependencies="{arg_name}")

    @c_wiring.wire
    class C:
        def f(self, x: Dummy):
            return x

    assert C().f() is world.get('x')

    d_wiring = Wiring(methods=['f'], dependencies=lambda arg: arg.name * 2)

    @d_wiring.wire
    class D:
        def f(self, x: Dummy):
            return x

    assert D().f() is world.get('xx')


def test_wiring_use_names():
    a_wiring = Wiring(methods=['f'])

    @a_wiring.wire
    class A:
        def f(self, y):
            return y

    with pytest.raises(TypeError):
        A().f()

    b_wiring = Wiring(methods=['f'], use_names=False)

    @b_wiring.wire
    class B:
        def f(self, y):
            return y

    with pytest.raises(TypeError):
        B().f()

    c_wiring = Wiring(methods=['f'], use_names=True)

    @c_wiring.wire
    class C:
        def f(self, y):
            return y

    assert C().f() is world.get('y')

    d_wiring = Wiring(methods=['f'], use_names=['y'])

    @d_wiring.wire
    class D:
        def f(self, y):
            return y

    assert D().f() is world.get('y')


def test_wiring_use_type_hints():
    a_wiring = Wiring(methods=['f'])

    @a_wiring.wire
    class A:
        def f(self, y: Dummy):
            return y

    assert A().f() is world.get(Dummy)

    b_wiring = Wiring(methods=['f'], use_type_hints=False)

    @b_wiring.wire
    class B:
        def f(self, y: Dummy):
            return y

    with pytest.raises(TypeError):
        B().f()

    c_wiring = Wiring(methods=['f'], use_type_hints=True)

    @c_wiring.wire
    class C:
        def f(self, y: Dummy):
            return y

    assert C().f() is world.get(Dummy)

    d_wiring = Wiring(methods=['f'], use_type_hints=['y'])

    @d_wiring.wire
    class D:
        def f(self, y: Dummy):
            return y

    assert D().f() is world.get(Dummy)


def test_complex_wiring():
    wiring = Wiring(methods=['f', 'g'],
                    use_names=['x'],
                    wire_super=True)

    class A:
        def g(self, x, y=None):
            return x, y

    @wiring.wire
    class B(A):
        def f(self, x: Dummy):
            return x

    assert B().f() is world.get(Dummy)
    assert B().g() == (world.get('x'), None)


def test_wiring_static_class_method():
    a_wiring = Wiring(methods=['static', 'klass'])

    @a_wiring.wire
    class A:
        @staticmethod
        def static(x: Dummy):
            return x

        @classmethod
        def klass(cls, x: Dummy):
            return cls, x

    assert A.static() is world.get(Dummy)
    assert A.klass() == (A, world.get(Dummy))
    assert A().static() is world.get(Dummy)
    assert A().klass() == (A, world.get(Dummy))

    class B:
        @staticmethod
        def static(x: Dummy):
            return x

        @classmethod
        def klass(cls, x: Dummy):
            return cls, x

    c_wiring = Wiring(methods=['static', 'klass'], wire_super=True)

    @c_wiring.wire
    class C(B):
        pass

    assert C.static() is world.get(Dummy)
    assert C.klass() == (C, world.get(Dummy))
    assert C().static() is world.get(Dummy)
    assert C().klass() == (C, world.get(Dummy))


def test_ignore_missing_method():
    with pytest.raises(AttributeError, match=".*unknown_method.*"):
        wiring = Wiring(methods=['unknown_method'])

        @wiring.wire
        class X:
            pass

    with pytest.raises(AttributeError, match=".*unknown_method.*"):
        wiring = Wiring(methods=['unknown_method'], ignore_missing_method=False)

        @wiring.wire
        class Y:
            pass

    a_wiring = Wiring(methods=['unknown_method'], ignore_missing_method=True)

    @a_wiring.wire
    class A:
        pass

    b_wiring = Wiring(methods=['method', 'maybe_method'],
                      ignore_missing_method=['maybe_method'])

    @b_wiring.wire
    class B:
        def method(self):
            pass

    with pytest.raises(AttributeError, match=".*method.*"):
        @b_wiring.wire
        class C:
            def maybe_method(self):
                pass


def test_unwrap_class_static_methods():
    wiring = Wiring(methods=['static', 'klass'],
                    use_names=True)  # use_names to force injection

    @wiring.wire
    class Dummy:
        @staticmethod
        def static(x):
            pass

        @classmethod
        def klass(cls, x):
            pass

    assert isinstance(Dummy.__dict__['static'], staticmethod)
    assert isinstance(Dummy.__dict__['klass'], classmethod)


@pytest.mark.parametrize('kwargs', [
    dict(methods={'method', 'test'}),
    dict(dependencies="{arg_name}"),
    dict(use_names=True),
    dict(use_type_hints=True),
    dict(wire_super={'method'}),
    dict(ignore_missing_method={'method'})
])
def test_copy(kwargs):
    wiring = Wiring(methods=['method'],
                    dependencies=dict(),
                    use_names=False,
                    use_type_hints=False,
                    wire_super=False,
                    ignore_missing_method=False)
    copy = wiring.copy(**kwargs)
    for key, value in kwargs.items():
        assert getattr(copy, key) == value


class DummyConf(WithWiringMixin):
    def __init__(self, wiring=None):
        self.wiring = wiring

    def copy(self, wiring):
        return DummyConf(wiring)


@pytest.mark.parametrize('kwargs', [
    dict(methods={'method', 'test'}),
    dict(dependencies="{arg_name}"),
    dict(use_names=True),
    dict(use_type_hints=True),
    dict(wire_super={'method'}),
    dict(ignore_missing_method={'method'})
])
def test_with_wiring(kwargs):
    conf = DummyConf(Wiring(methods=['method'],
                            dependencies=dict(),
                            use_names=False,
                            use_type_hints=False,
                            wire_super=False,
                            ignore_missing_method=False))
    copy = conf.with_wiring(**kwargs)
    for key, value in kwargs.items():
        assert getattr(copy.wiring, key) == value

    kwargs.setdefault('methods', {'method'})
    copy = DummyConf().with_wiring(**kwargs)
    for key, value in kwargs.items():
        assert getattr(copy.wiring, key) == value


def test_invalid_with_wiring():
    with pytest.raises(TypeError):
        DummyConf().with_wiring()



import pytest

from antidote import world
from antidote._internal.utils import short_id
from antidote._providers import Tag, Tagged, TagProvider
from antidote.core import DependencyInstance
from antidote.exceptions import (DependencyNotFoundError, DuplicateTagError,
                                 FrozenWorldError)


class Service:
    pass


@pytest.fixture()
def provider():
    with world.test.empty():
        world.provider(TagProvider)
        yield world.get(TagProvider)


def test_tag():
    # id is unique
    assert Tag().group != Tag().group
    assert repr(Tag()) != repr(Tag())

    # Immutable
    with pytest.raises(AttributeError):
        Tag().x = 1


def test_custom_tag():
    class CustomTag(Tag):
        def group(self):
            return "hello"

    assert "hello" in repr(CustomTag())


def test_repr(provider: TagProvider):
    class A:
        pass

    tag = Tag()
    provider.register(A, tags=[tag])

    assert str(tag) in repr(provider)
    assert str(A) in repr(provider)


def test_tagged_dependencies():
    with world.test.empty():
        world.singletons.add({
            'd': object(),
            'd2': object()
        })
        tag = Tag()

        from antidote._internal.state import get_container
        t = Tagged(
            container=get_container(),
            dependencies=['d', 'd2'],
            tags=[tag, tag]
        )

        assert set(t.tags()) == {tag}
        assert len(t) == 2

        expected = [world.get('d'), world.get('d2')]
        assert list(t.values()) == expected

        # tagged dependencies should only be retrieved once.
        with world.test.empty():
            world.singletons.add({
                'd': object(),
                'd2': object()
            })
            assert list(t.values()) == expected


def test_tagged_dependencies_invalid_dependency():
    with world.test.empty():
        tag = Tag()

        from antidote._internal.state import get_container
        t = Tagged(
            container=get_container(),
            dependencies=['d'],
            tags=[tag]
        )
        assert set(t.tags()) == {tag}
        assert len(t) == 1

        with pytest.raises(DependencyNotFoundError):
            list(t.values())


def test_exists():
    provider = TagProvider()
    assert not provider.exists(Tag())

    tag = Tag()
    provider.register('test', tags=[tag])
    assert provider.exists(tag)

    class CustomTag(Tag):
        def group(self):
            return 'group'

    assert not provider.exists(CustomTag())
    provider.register('test2', tags=[CustomTag()])
    assert provider.exists(CustomTag())


def test_provide_unknown_tag():
    provider = TagProvider()
    assert world.test.maybe_provide_from(provider, Tag()) is None


def test_provide_tags():
    with world.test.empty():
        provider = TagProvider()
        world.singletons.add(dict(test=object(), test2=object()))
        tagA = Tag()
        tagB = Tag()
        provider.register('test', tags=[tagA, tagB])
        provider.register('test2', tags=[tagB])

        result = world.test.maybe_provide_from(provider, tagA)
        assert isinstance(result, DependencyInstance)
        assert result.singleton is False
        tagged_dependencies: Tagged = result.value
        assert len(tagged_dependencies) == 1
        assert set(tagged_dependencies.tags()) == {tagA}
        assert {world.get('test')} == set(tagged_dependencies.values())

        result = world.test.maybe_provide_from(provider, tagB)
        assert isinstance(result, DependencyInstance)
        assert result.singleton is False
        tagged_dependencies: Tagged = result.value
        assert len(tagged_dependencies) == 2
        assert {(tagB, world.get('test')),
                (tagB, world.get('test2'))} == set(tagged_dependencies.items())
        assert set(tagged_dependencies.tags()) == {tagB}
        assert {world.get('test'), world.get('test2')} == set(
            tagged_dependencies.values())


def test_custom_tags(provider: TagProvider):
    class CustomTag(Tag):
        __slots__ = ('name', 'attr')
        name: str
        attr: int

        def __init__(self, name: str, attr: int = 0):
            super().__init__(name=name, attr=attr)

        def group(self):
            return self.name

    world.singletons.add(dict(test=object(), test2=object()))
    test_tags = [CustomTag("A", attr=1), CustomTag("B", attr=1)]
    provider.register('test', tags=test_tags)
    test2_tags = [CustomTag("B", attr=2)]
    provider.register('test2', tags=test2_tags)

    result = world.get(CustomTag("A"))
    assert list(result.tags()) == [test_tags[0]]
    assert list(result.values()) == [world.get('test')]

    result = world.get(CustomTag("B"))
    if list(result.tags()) == [test_tags[1], test2_tags[0]]:
        assert list(result.values()) == [world.get('test'), world.get('test2')]
    elif list(result.tags()) == [test2_tags[0], test_tags[1]]:
        assert list(result.values()) == [world.get('test2'), world.get('test')]
    else:
        assert False


def test_duplicate_tag_error(provider: TagProvider):
    tag = Tag()

    with world.test.clone():
        provider = world.get(TagProvider)
        provider.register('test', tags=[tag])
        with pytest.raises(DuplicateTagError) as exc_info:
            provider.register('test', tags=[tag])
        # short_id being base64 can contain the chars '+/' which interfere with regex
        # matching.
        assert short_id(tag) in str(exc_info.value)

    with world.test.clone():
        provider = world.get(TagProvider)
        with pytest.raises(DuplicateTagError) as exc_info:
            provider.register('test', tags=[tag, tag])
        assert short_id(tag) in str(exc_info.value)


@pytest.mark.parametrize(
    'tags',
    [
        [object],
        [lambda _: False],
        ['test', object],
    ]
)
def test_invalid_tags(provider: TagProvider, tags):
    with pytest.raises(TypeError):
        provider.register('test', tags=tags)


def test_all_must_be_valid_for_tag():
    provider = TagProvider()
    tag = Tag()
    with pytest.raises(TypeError):
        provider.register('test', tags=[tag, 3])

    assert world.test.maybe_provide_from(provider, tag) is None


@pytest.mark.parametrize('dependency', ['test', Service, object()])
def test_unknown_dependency(dependency):
    provider = TagProvider()
    assert world.test.maybe_provide_from(provider, dependency) is None


@pytest.mark.parametrize('keep_singletons_cache', [True, False])
def test_copy(provider: TagProvider,
              keep_singletons_cache: bool):
    world.singletons.add(dict(test=object(), test2=object(), test3=object()))
    tag = Tag()
    provider.register('test', tags=[tag])
    cloned = provider.clone(keep_singletons_cache=keep_singletons_cache)

    with pytest.raises(DuplicateTagError):
        cloned.register('test', tags=[tag])

    assert list(world.test.maybe_provide_from(cloned, tag).value.tags()) == [tag]
    assert list(world.test.maybe_provide_from(cloned, tag).value.values()) == [
        world.get('test')]

    tag2 = Tag()
    provider.register('test2', tags=[tag2])
    assert world.test.maybe_provide_from(cloned, tag2) is None

    tag3 = Tag()
    cloned.register('test3', tags=[tag3])
    with pytest.raises(DependencyNotFoundError):
        world.get(tag3)


def test_freeze(provider: TagProvider):
    world.freeze()

    with pytest.raises(FrozenWorldError):
        provider.register("test", tags=[Tag()])



import enum

from .debug import debug_repr, short_id
from .immutable import FinalImmutable, Immutable
from .meta import AbstractMeta, FinalMeta
from .slots import SlotRecord
from .. import API


@API.private
class Copy(enum.Enum):
    IDENTICAL = object()


@API.private
def raw_getattr(cls: type, attr: str, with_super: bool = False):
    """
    Used to retrieve the 'raw' attribute from a class, typically the descriptor itself
    and not whatever it outputs.

    Args:
        cls: Class from which an attribute must be retrieved
        attr: Name of the attribute
        with_super: Whether the attribute must be search in the whole inheritance
        hierarchy or not.

    Returns:
        The attribute or raises a KeyError.
    """
    if with_super:
        for c in cls.__mro__:
            try:
                return c.__dict__[attr]
            except KeyError:
                pass

        raise AttributeError(f"{cls} has no attribute '{attr}'")
    else:
        try:
            return cls.__dict__[attr]
        except KeyError:
            pass
        raise AttributeError(
            f"Attribute '{attr}' is not defined in {cls}. "
            f"Mother classes are not taken into account. "
            f"This may indicate that 'wire_super' is not configured properly.")



import threading
from collections import deque
from contextlib import contextmanager
from typing import (Any, Callable, Deque, Dict, Hashable, List, Mapping, Optional,
                    Tuple, Type)
from weakref import ref

# @formatter:off
cimport cython
from cpython.mem cimport PyMem_Free, PyMem_Malloc
from cpython.ref cimport PyObject
from fastrlock.rlock cimport create_fastrlock, lock_fastrlock, unlock_fastrlock

from antidote._internal.stack cimport DependencyStack
from .exceptions import (DependencyCycleError, DependencyInstantiationError,
                         DependencyNotFoundError, DuplicateDependencyError, FrozenWorldError)
# @formatter:on
from .utils import DependencyDebug


cdef extern from "Python.h":
    PyObject*PyList_GET_ITEM(PyObject *list, Py_ssize_t i)
    Py_ssize_t PyList_Size(PyObject *list)
    Py_ssize_t PyTuple_GET_SIZE(PyObject *p)
    int PyDict_SetItem(PyObject *p, PyObject *key, PyObject *val) except -1
    PyObject*PyDict_GetItem(PyObject *p, PyObject *key)


FLAG_DEFINED = 1
FLAG_SINGLETON = 2


@cython.freelist(64)
@cython.final
cdef class DependencyInstance:
    def __cinit__(self, value, *, bint singleton = False):
        self.value = value
        self.singleton = singleton

    def __repr__(self):
        return f"DependencyInstance(value={self.value}, singleton={self.singleton})"

    def __eq__(self, other):
        return isinstance(other, DependencyInstance) \
               and self.singleton == other.singleton \
               and self.value == other.value


@cython.freelist(64)
cdef class PyObjectBox:
    """
    Used to hold a Python object for DependencyResult. It is NOT initialized.
    """

cdef class Container:
    def get(self, dependency: Hashable):
        raise NotImplementedError()  # pragma: no cover

    def provide(self, dependency: Hashable):
        raise NotImplementedError()  # pragma: no cover

cdef class RawProvider:
    """
    Contrary to the Python implementation, the container does not rely on provide() but
    on fast_provide(). This is done to improve performance by avoiding object creation.
    """

    def __init__(self):
        self._container_ref = None

    def clone(self, keep_singletons_cache: bool) -> 'RawProvider':
        raise NotImplementedError()  # pragma: no cover

    def exists(self, dependency: Hashable) -> bool:
        raise NotImplementedError()  # pragma: no cover

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        raise NotImplementedError()  # pragma: no cover

    def maybe_provide(self,
                      dependency: Hashable,
                      container: Container) -> DependencyInstance:
        raise NotImplementedError()

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        cdef:
            DependencyInstance dependency_instance
        dependency_instance = self.maybe_provide(<object> dependency,
                                                 <Container> container)
        if dependency_instance is not None:
            result.flags = FLAG_DEFINED | (
                FLAG_SINGLETON if dependency_instance.singleton else 0)
            (<PyObjectBox> result.box).obj = dependency_instance.value

    @property
    def is_registered(self):
        return self._container_ref is not None

    @contextmanager
    def _ensure_not_frozen(self):
        if self._container_ref is None:
            yield
        else:
            container = self._container_ref()
            assert container is not None, "Associated container does not exist anymore."
            with container.ensure_not_frozen():
                yield

    def _raise_if_exists(self, dependency):
        if self._container_ref is not None:
            container: RawContainer = self._container_ref()
            assert container is not None, "Associated container does not exist anymore."
            container.raise_if_exists(dependency)
        else:
            if self.exists(dependency):
                raise DuplicateDependencyError(
                    f"{dependency} has already been registered in {type(self)}")

cdef class FastProvider(RawProvider):
    def maybe_provide(self,
                      dependency: Hashable,
                      container: Container) -> DependencyInstance:
        cdef:
            DependencyResult result
            PyObjectBox box = PyObjectBox.__new__(PyObjectBox)
        result.box = <PyObject*> box

        result.flags = 0
        self.fast_provide(<PyObject*> dependency, <PyObject*> container, &result)
        if result.flags != 0:
            return DependencyInstance(box.obj,
                                      singleton=(result.flags & FLAG_SINGLETON) != 0)
        return None

    cdef fast_provide(self,
                      PyObject*dependency,
                      PyObject*container,
                      DependencyResult*result):
        raise NotImplementedError()

cdef class RawContainer(Container):
    """
    Behaves the same as the Python implementation but with additional optimizations:

    - singletons clock: Avoid testing twice the singleton dictionary it hasn't changed
                        since the first, not thread-safe, check.
    - cache: Recurring non singletons dependencies will not go through the usual _providers
             loop and their provider will be cached. This avoids the overhead of each
             provider checking whether it's a dependency it can provide or not.
    """

    def __init__(self, *, cache_size: int = 16):
        cdef:
            size_t capacity
            size_t*counter
        if not isinstance(cache_size, int):
            raise TypeError("cache_size must be a strictly positive integer.")
        if cache_size <= 0:
            raise ValueError("cache_size must be a strictly positive integer.")
        capacity = <size_t> cache_size

        self._providers = list()  # type: List[RawProvider]
        self._singletons = dict()  # type: Dict[Any, Any]
        self._dependency_stack = DependencyStack()
        self._singleton_lock = create_fastrlock()
        self.__freeze_lock = threading.RLock()
        self.__frozen = False

        # Cython optimizations
        self._singletons_clock = 0
        self.__cache.length = 0
        self.__cache.capacity = capacity
        self.__cache.dependencies = <PyObject**> PyMem_Malloc(
            capacity * sizeof(PyObject*))
        self.__cache.providers = <PyObject**> PyMem_Malloc(capacity * sizeof(PyObject*))
        self.__cache.counters = <size_t*> PyMem_Malloc(capacity * sizeof(size_t))
        if not self.__cache.dependencies or not self.__cache.providers or not self.__cache.counters:
            raise MemoryError()
        for counter in self.__cache.counters[:capacity]:
            counter[0] = 0

    def __dealloc__(self):
        PyMem_Free(self.__cache.dependencies)
        PyMem_Free(self.__cache.providers)
        PyMem_Free(self.__cache.counters)

    def __repr__(self):
        return f"{type(self).__name__}(_providers={', '.join(map(str, self._providers))})"

    @property
    def providers(self):
        return self._providers.copy()

    @contextmanager
    def ensure_not_frozen(self):
        with self.__freeze_lock:
            if self.__frozen:
                raise FrozenWorldError()
            yield

    def freeze(self):
        with self.__freeze_lock:
            self.__frozen = True

    def add_provider(self, provider_cls: Type[RawProvider]):
        cdef:
            RawProvider provider

        if not isinstance(provider_cls, type) \
                or not issubclass(provider_cls, RawProvider):
            raise TypeError(
                f"provider must be a Provider, not a {provider_cls}")

        with self.ensure_not_frozen(), self._singleton_lock:
            if any(provider_cls == type(p) for p in self._providers):
                raise ValueError(f"Provider {provider_cls} already exists")

            provider = provider_cls()
            provider._container_ref = ref(self)
            self._providers.append(provider)
            self._singletons[provider_cls] = provider
            self._singletons_clock += 1

    def add_singletons(self, dependencies: Mapping):
        with self.ensure_not_frozen(), self._singleton_lock:
            for k, v in dependencies.items():
                self.raise_if_exists(k)
            self._singletons.update(dependencies)
            self._singletons_clock += 1

    def raise_if_exists(self, dependency: Hashable):
        with self.__freeze_lock:
            if dependency in self._singletons:
                raise DuplicateDependencyError(
                    f"{dependency!r} has already been defined as a singleton pointing "
                    f"to {self._singletons[dependency]}")

            for provider in self._providers:
                if provider.exists(dependency):
                    debug = provider.maybe_debug(dependency)
                    message = f"{dependency!r} has already been declared " \
                              f"in {type(provider)}"
                    if debug is None:
                        raise DuplicateDependencyError(message)
                    else:
                        raise DuplicateDependencyError(message + f"\n{debug.info}")

    def clone(self,
              *,
              keep_singletons: bool = False,
              clone_providers: bool = True) -> 'RawContainer':
        cdef:
            RawProvider clone
            RawProvider p
            RawContainer c = type(self)()

        with self._singleton_lock:
            if keep_singletons:
                c._singletons = self._singletons.copy()
            if clone_providers:
                for p in self._providers:
                    clone = p.clone(keep_singletons)
                    if clone is p or clone._container_ref is not None:
                        raise RuntimeError("A Provider should always return a fresh "
                                           "instance when copy() is called.")
                    clone._container_ref = ref(c)
                    c._providers.append(clone)
                    c._singletons[type(p)] = clone
            else:
                for p in self._providers:
                    c.add_provider(type(p))

        return c

    def debug(self, dependency: Hashable) -> DependencyDebug:
        from .._internal.utils.debug import debug_repr

        for p in self._providers:
            debug = p.maybe_debug(dependency)
            if debug is not None:
                return debug
        try:
            value = self._singletons[dependency]
            return DependencyDebug(f"Singleton {debug_repr(dependency)} "
                                   f"-> {value!r}",
                                   singleton=True)
        except KeyError:
            raise DependencyNotFoundError(dependency)

    def provide(self, dependency: Hashable):
        cdef:
            DependencyResult result
            PyObjectBox box = PyObjectBox.__new__(PyObjectBox)
        result.box = <PyObject*> box

        self.fast_get(<PyObject*> dependency, &result)
        if result.flags != 0:
            return DependencyInstance(box.obj,
                                      singleton=(result.flags & FLAG_SINGLETON) != 0)
        raise DependencyNotFoundError(dependency)

    def get(self, dependency: Hashable):
        cdef:
            DependencyResult result
            PyObjectBox box = PyObjectBox.__new__(PyObjectBox)
        result.box = <PyObject*> box

        self.fast_get(<PyObject*> dependency, &result)
        if result.flags != 0:
            return box.obj
        raise DependencyNotFoundError(dependency)

    # No ownership from here on. You MUST keep a valid reference to dependency.
    cdef fast_get(self, PyObject*dependency, DependencyResult*result):
        cdef:
            PyObject*ptr
            PyObjectBox box
            unsigned long clock = self._singletons_clock

        result.flags = 0
        ptr = PyDict_GetItem(<PyObject*> self._singletons, dependency)
        if ptr != NULL:
            result.flags = FLAG_DEFINED | FLAG_SINGLETON
            (<PyObjectBox> result.box).obj = <object> ptr
        else:
            self.__safe_provide(dependency, result, clock)

    cdef __safe_provide(self,
                        PyObject*dependency,
                        DependencyResult*result,
                        unsigned long singletons_clock):
        cdef:
            PyObject*ptr
            Exception error
            object lock = self._singleton_lock
            PyObject*singletons = <PyObject*> self._singletons
            PyObject*stack = <PyObject*> self._dependency_stack
            ProviderCache*cache
            PyObject** cached_dependency
            PyObject*providers
            size_t i

        lock_fastrlock(lock, -1, True)

        # If anything changed in the singletons the clock would be different
        # otherwise no need to re-check the dictionary.
        if singletons_clock < self._singletons_clock:
            ptr = PyDict_GetItem(singletons, dependency)
            if ptr != NULL:
                unlock_fastrlock(lock)
                result.flags = FLAG_DEFINED | FLAG_SINGLETON
                (<PyObjectBox> result.box).obj = <object> ptr
                return

        if 0 != (<DependencyStack> stack).push(dependency):
            error = (<DependencyStack> stack).reset_with_error(dependency)
            unlock_fastrlock(lock)
            raise error

        try:
            cache = &self.__cache
            i = 0
            for cached_dependency in cache.dependencies[:cache.length]:
                if dependency == cached_dependency[0]:
                    (<RawProvider> cache.providers[i]).fast_provide(
                        dependency, <PyObject*> self, result)
                    if result.flags != 0:
                        if (result.flags & FLAG_SINGLETON) != 0:
                            PyDict_SetItem(singletons,
                                           dependency,
                                           <PyObject*> (<PyObjectBox> result.box).obj)
                            self._singletons_clock += 1
                        elif i > 0:
                            cache_update(cache, i)
                        else:
                            cache.counters[i] += 1
                    return
                i += 1

            providers = <PyObject*> self._providers
            for i in range(<size_t> PyList_Size(providers)):
                (<RawProvider> PyList_GET_ITEM(providers, i)).fast_provide(
                    dependency,
                    <PyObject*> self,
                    result
                )
                if result.flags != 0:
                    if (result.flags & FLAG_SINGLETON) != 0:
                        PyDict_SetItem(singletons,
                                       dependency,
                                       <PyObject*> (<PyObjectBox> result.box).obj)
                        self._singletons_clock += 1
                    else:
                        if cache.length < cache.capacity:
                            cache.length += 1
                        ptr = PyList_GET_ITEM(providers, i)
                        i = cache.length - 1
                        cache.providers[i] = ptr
                        cache.dependencies[i] = dependency
                        cache_update(cache, i)
                    return

        except Exception as error:
            if isinstance(error, DependencyCycleError):
                raise
            raise DependencyInstantiationError(<object> dependency) from error
        finally:
            (<DependencyStack> stack).pop()
            unlock_fastrlock(lock)

# Imitating the SpaceSaving algorithm
cdef inline void cache_update(ProviderCache*cache, size_t pos) nogil:
    cdef:
        size_t counter, i = pos, new_pos = 0
        PyObject*provider
        PyObject*dependency
    counter = cache.counters[pos] + 1
    while 0 < i:
        i -= 1
        if counter < cache.counters[i]:
            new_pos = i + 1
            break

    if pos != new_pos:
        dependency = cache.dependencies[pos]
        provider = cache.providers[pos]
        cache.dependencies[pos] = cache.dependencies[new_pos]
        cache.counters[pos] = cache.counters[new_pos]
        cache.providers[pos] = cache.providers[new_pos]
        cache.dependencies[new_pos] = dependency
        cache.providers[new_pos] = provider
    cache.counters[new_pos] = counter


cdef class OverridableRawContainer(RawContainer):
    cdef:
        object __override_lock
        dict __singletons_override
        dict __factory_overrides
        object __provider_overrides


    def __init__(self):
        super().__init__()
        self.__override_lock = threading.RLock()
        # Used to differentiate singletons from the overrides and the "normal" ones.
        self.__singletons_override = dict()
        self.__factory_overrides: Dict[Any, Tuple[Callable[[], Any], bool]] = {}
        self.__provider_overrides: Deque[
            Callable[[Any], Optional[DependencyInstance]]] = deque()

    @classmethod
    def build(cls,
              original: RawContainer,
              keep_singletons: bool) -> 'OverridableRawContainer':
        cdef:
            RawContainer clone = original.clone(keep_singletons=keep_singletons)
            OverridableRawContainer container = OverridableRawContainer()

        container._singletons = clone._singletons
        container._providers = clone._providers
        if isinstance(clone, OverridableRawContainer):
            container.__singletons_override = (<OverridableRawContainer> clone).__singletons_override
            container.__factory_overrides = (<OverridableRawContainer> clone).__factory_overrides
            container.__provider_overrides = (<OverridableRawContainer> clone).__provider_overrides
        return container

    def override_singletons(self, singletons: dict):
        if not isinstance(singletons, dict):
            raise TypeError(f"singletons must be a dict, not a {type(singletons)}")
        with self.__override_lock:
            self.__singletons_override.update(singletons)

    def override_factory(self,
                         dependency: Hashable,
                         *,
                         factory: Callable[[], Any],
                         singleton: bool):
        if not callable(factory):
            raise TypeError(f"factory must be a callable, not a {type(factory)}")
        if not isinstance(singleton, bool):
            raise TypeError(f"singleton must be a boolean, not a {type(singleton)}")
        with self.__override_lock:
            self.__factory_overrides[dependency] = (factory, singleton)

    def override_provider(self,
                          provider: Callable[[Any], Optional[DependencyInstance]]):
        if not callable(provider):
            raise TypeError(f"provider must be a callable, not a {type(provider)}")
        with self.__override_lock:
            self.__provider_overrides.appendleft(provider)  # latest provider wins

    def clone(self,
              *,
              keep_singletons: bool = False,
              clone_providers: bool = True) -> 'OverridableRawContainer':
        cdef:
            OverridableRawContainer container

        with self._singleton_lock, self.__override_lock:
            container = super().clone(keep_singletons=keep_singletons,
                                      clone_providers=clone_providers)
            if keep_singletons:
                container.__singletons_override = self.__singletons_override.copy()
            container.__factory_overrides = self.__factory_overrides.copy()
            container.__provider_overrides = self.__provider_overrides.copy()

        return container

    def debug(self, dependency: Hashable) -> DependencyDebug:
        from .._internal.utils.debug import debug_repr

        with self.__override_lock:
            try:
                value = self.__singletons_override[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Singleton: {debug_repr(dependency)} "
                                       f"-> {value!r}",
                                       singleton=True)
            try:
                (factory, singleton) = self.__factory_overrides[dependency]
            except KeyError:
                pass
            else:
                return DependencyDebug(f"Override/Factory: {debug_repr(dependency)} "
                                       f"-> {debug_repr(factory)}",
                                       singleton=singleton)

        return super().debug(dependency)

    # Not really fast anymore, but we don't really care in tests.
    cdef fast_get(self, PyObject*dependency, DependencyResult*result):
        cdef:
            PyObject*ptr
            PyObjectBox box

        dep = <object> dependency
        result.flags = 0
        with self.__override_lock, self._singleton_lock:
            with self._dependency_stack.instantiating(dep):
                try:
                    obj = self.__singletons_override[dep]
                except KeyError:
                    pass
                else:
                    result.flags = FLAG_DEFINED | FLAG_SINGLETON
                    (<PyObjectBox> result.box).obj = obj
                    return

                try:
                    for provider in self.__provider_overrides:
                        dependency_instance = provider(dep)
                        if dependency_instance is not None:
                            if dependency_instance.singleton:
                                self.__singletons_override[dep] = dependency_instance.value
                                result.flags = FLAG_DEFINED | FLAG_SINGLETON
                            else:
                                result.flags = FLAG_DEFINED
                            (<PyObjectBox> result.box).obj = dependency_instance.value
                            return

                    try:
                        (factory, singleton) = self.__factory_overrides[dep]
                    except KeyError:
                        pass
                    else:
                        obj = factory()
                        if singleton:
                            self.__singletons_override[dep] = obj
                            result.flags = FLAG_DEFINED | FLAG_SINGLETON
                        else:
                            result.flags = FLAG_DEFINED
                        (<PyObjectBox> result.box).obj = obj
                        return

                except Exception as error:
                    if isinstance(error, DependencyCycleError):
                        raise
                    raise DependencyInstantiationError(<object> dependency) from error

            RawContainer.fast_get(self, dependency, result)



from typing import Callable, Dict, Hashable, Optional

from .._internal import API
from .._internal.utils import debug_repr, FinalImmutable
from ..core import Container, DependencyInstance, Provider
from ..core.utils import DependencyDebug


@API.private
class IndirectProvider(Provider):
    def __init__(self):
        super().__init__()
        self.__links: Dict[Hashable, Link] = dict()
        self.__static_links: Dict[Hashable, Hashable] = dict()

    def __repr__(self):
        return f"{type(self).__name__}(links={self.__links}, " \
               f"static_links={self.__static_links})"

    def clone(self, keep_singletons_cache: bool) -> 'IndirectProvider':
        p = IndirectProvider()
        p.__links = self.__links.copy()
        p.__static_links = self.__static_links.copy()
        return p

    def exists(self, dependency: Hashable) -> bool:
        return dependency in self.__static_links or dependency in self.__links

    def maybe_debug(self, dependency: Hashable) -> Optional[DependencyDebug]:
        try:
            link = self.__links[dependency]
        except KeyError:
            pass
        else:
            repr_d = debug_repr(dependency)
            linker = link.linker  # type: ignore  # Mypy treats linker as a method
            repr_linker = debug_repr(linker)
            if link.permanent:
                if dependency in self.__static_links:
                    target = self.__static_links[dependency]
                    return DependencyDebug(
                        f"Permanent link: {repr_d} -> {debug_repr(target)} "
                        f"defined by {repr_linker}",
                        singleton=True,
                        dependencies=[target])
                else:
                    return DependencyDebug(
                        f"Permanent link: {repr_d} -> ??? "
                        f"defined by {repr_linker}",
                        singleton=True)
            else:
                return DependencyDebug(
                    f"Dynamic link: {repr_d} -> ??? defined by {repr_linker}",
                    singleton=False,
                    wired=[linker])

        try:
            target = self.__static_links[dependency]
        except KeyError:
            pass
        else:
            repr_d = debug_repr(dependency)
            return DependencyDebug(f"Static link: {repr_d} -> {debug_repr(target)}",
                                   singleton=True,
                                   dependencies=[target])
        return None

    def maybe_provide(self, dependency: Hashable, container: Container
                      ) -> Optional[DependencyInstance]:
        try:
            target = self.__static_links[dependency]
        except KeyError:
            pass
        else:
            return container.provide(target)

        try:
            link = self.__links[dependency]
        except KeyError:
            pass
        else:
            target = link.linker()  # type: ignore  # Mypy treats linker as a method
            if link.permanent:
                self.__static_links[dependency] = target
            t = container.provide(target)
            return DependencyInstance(
                t.value,
                singleton=t.singleton and link.permanent
            )

        return None

    def register_static(self, dependency: Hashable, target_dependency: Hashable):
        self._assert_not_duplicate(dependency)
        self.__static_links[dependency] = target_dependency

    def register_link(self, dependency: Hashable, linker: Callable[[], Hashable],
                      permanent: bool = True):
        self._assert_not_duplicate(dependency)
        self.__links[dependency] = Link(linker, permanent)


@API.private
class Link(FinalImmutable):
    __slots__ = ('linker', 'permanent')
    linker: Callable[[], Hashable]
    permanent: bool



import collections.abc as c_abc
from typing import (Any, Callable, Hashable, Iterable, Mapping,
                    Optional, overload, Sequence, TypeVar, Union)

from ._injection import raw_inject
from .._compatibility.typing import final
from .._internal import API
from .._internal.utils import FinalImmutable

F = TypeVar('F', Callable, staticmethod, classmethod)


@API.public
@final
class Arg(FinalImmutable):
    """
    Represents an argument (name and type hint) if you need a very custom injection
    logic.
    """
    __slots__ = ('name', 'type_hint')
    name: str
    type_hint: Any

    def __init__(self, name: str, type_hint: Any):
        super().__init__(name=name, type_hint=type_hint)


# This type is experimental.
DEPENDENCIES_TYPE = Union[
    Mapping[str, Hashable],  # {arg_name: dependency, ...}
    Sequence[Optional[Hashable]],  # (dependency for arg 1, ...)
    Callable[[Arg], Optional[Hashable]],  # arg -> dependency
    str  # str.format(arg_name=arg_name) -> dependency
]


@overload
def inject(func: F,  # noqa: E704  # pragma: no cover
           *,
           dependencies: DEPENDENCIES_TYPE = None,
           use_names: Union[bool, Iterable[str]] = None,
           use_type_hints: Union[bool, Iterable[str]] = None
           ) -> F: ...


@overload
def inject(*,  # noqa: E704  # pragma: no cover
           dependencies: DEPENDENCIES_TYPE = None,
           use_names: Union[bool, Iterable[str]] = None,
           use_type_hints: Union[bool, Iterable[str]] = None
           ) -> Callable[[F], F]: ...


@API.public
def inject(func=None,
           *,
           dependencies: DEPENDENCIES_TYPE = None,
           use_names: Union[bool, Iterable[str]] = None,
           use_type_hints: Union[bool, Iterable[str]] = None):
    """
    Inject the dependencies into the function lazily, they are only retrieved
    upon execution. As several options can apply to the same argument, the priority is
    defined as:

    1. Dependencies declared explicitly if any declared with :code:`dependencies`.
    2. Type hints (unless deactivated through :code:`use_type_hints`)
    3. Argument names if specified with :code:`use_names`

    .. doctest:: core_inject

        >>> from antidote import inject
        >>> # All possibilities for dependency argument.
        ... @inject(dependencies=None)  # default
        ... def f(a):
        ...     pass  # a = <not injected>
        >>> @inject(dependencies=dict(b='dependency'))
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> @inject(dependencies=[None, 'dependency'])  # Nothing will be injected a
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> @inject(dependencies=lambda arg: 'dependency' if arg.name == 'b' else None)
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('dependency')
        >>> @inject(dependencies="conf:{arg_name}")
        ... def f(a):
        ...     pass  # a = world.get('conf:a')
        >>> # All possibilities for use_names argument.
        ... @inject(use_names=False)  # default
        ... def f(a):
        ...     pass  # a = <not injected>
        >>> @inject(use_names=True)
        ... def f(a):
        ...     pass  # a = world.get('a')
        >>> @inject(use_names=['b'])
        ... def f(a, b):
        ...     pass  # a, b = <not injected>, world.get('b')
        >>> # All possibilities for use_type_hints argument.
        ... class Service:
        ...     pass
        >>> @inject(use_type_hints=True)  # default
        ... def f(a: Service):
        ...     pass  # a = world.get(Service)
        >>> @inject(use_type_hints=['b'])
        ... def f(a: Service, b: Service):
        ...     pass  # a, b = <not injected>, world.get(Service)
        >>> @inject(use_type_hints=False)
        ... def f(a: Service):
        ...     pass  # a = <not injected>

    Args:
        func: Callable to be wrapped. Can also be used on static methods or class methods.
        dependencies: Explicit definition of the dependencies which overrides
            :code:`use_names` and :code:`use_type_hints`. Defaults to :py:obj:`None`.
            Can be one of:

            - Mapping from argument name to its dependency
            - Sequence of dependencies which will be mapped with the position
              of the arguments. :py:obj:`None` can be used as a placeholder.
            - Callable which receives :py:class:`~.Arg` as arguments and should
              return the matching dependency. :py:obj:`None` should be used for
              arguments without dependency.
            - String which must have :code:`{arg_name}` as format parameter
        use_names: Whether or not the arguments' name should be used as their
            respective dependency. An iterable of argument names may also be
            supplied to activate this feature only for those. Defaults to :code:`False`.
        use_type_hints: Whether or not the type hints should be used as the arguments
            dependency. An iterable of argument names may also be supplied to activate
            this feature only for those. Any type hints from the builtins (str, int...)
            or the typing (except :py:class:`~typing.Optional`) are ignored. It overrides
            :code:`use_names`. Defaults to :code:`True`.

    Returns:
        The decorator to be applied or the injected function if the
        argument :code:`func` was supplied.

    """
    return raw_inject(func,
                      dependencies=dependencies,
                      use_names=use_names,
                      use_type_hints=use_type_hints)


@API.experimental  # Function will be kept in sync with @inject, so you may use it.
def validate_injection(dependencies: DEPENDENCIES_TYPE = None,
                       use_names: Union[bool, Iterable[str]] = None,
                       use_type_hints: Union[bool, Iterable[str]] = None):
    if not (dependencies is None
            or isinstance(dependencies, (str, c_abc.Sequence, c_abc.Mapping))
            or callable(dependencies)):
        raise TypeError(
            f"dependencies can be None, a mapping of arguments names to dependencies, "
            f"a sequence of dependencies, a function or a string, "
            f"not a {type(dependencies)!r}"
        )
    if isinstance(dependencies, str):
        if "{arg_name}" not in dependencies:
            raise ValueError("Missing formatting parameter {arg_name} in dependencies. "
                             "If you really want a constant injection, "
                             "consider using a defaultdict.")
    if isinstance(dependencies, c_abc.Mapping):
        if not all(isinstance(k, str) for k in dependencies.keys()):
            raise TypeError("Dependencies keys must be argument names (str)")

    if not (use_names is None or isinstance(use_names, (bool, c_abc.Iterable))):
        raise TypeError(
            f"use_names must be either a boolean or a whitelist of argument names, "
            f"not {type(use_names)!r}.")
    if isinstance(use_names, c_abc.Iterable):
        if not all(isinstance(arg_name, str) for arg_name in use_names):
            raise TypeError("use_names must be list of argument names (str) or a boolean")

    if not (use_type_hints is None or isinstance(use_type_hints,
                                                 (bool, c_abc.Iterable))):
        raise TypeError(
            f"use_type_hints must be either a boolean or a whitelist of argument names, "
            f"not {type(use_type_hints)!r}.")
    if isinstance(use_type_hints, c_abc.Iterable):
        if not all(isinstance(arg_name, str) for arg_name in use_type_hints):
            raise TypeError(
                "use_type_hints must be list of argument names (str) or a boolean")



"""
Utilities to overrides dependencies in Antidote. Those can only be used in a test world
created by :py:func:`.world.test.clone` with :code:`overridable=True`. Dependencies
can be overridden multiple times, it won't raise any error. Overridden dependencies also
won't be taken into account when Antidote checks for duplicates. Hence you may do the
following:

.. doctest:: world_override

    >>> from antidote import world
    >>> with world.test.clone(overridable=True):
    ...     world.test.override.singleton('test', 1)
    ...     # Override a second time a singleton
    ...     world.test.override.singleton('test', 1)
    ...     # Declaring normally the same singleton afterwards
    ...     # won't raise any errors.
    ...     world.singletons.add('test', 1)

.. note::

    Overrides works as a layer on top of the usual dependencies. So while they don't
    interfer with latter, they can interfer with each other.

    .. doctest:: world_override

        >>> from antidote import world
        >>> with world.test.clone(overridable=True):
        ...     world.test.override.singleton('test', 1)
        ...     # 'test' is already a singleton in the override layer, so the factory
        ...     # won't be taken into account.
        ...     @world.test.override.factory('test')
        ...     def test():
        ...         return 2
        ...     world.get[int]("test")
        1

    The priority of the overrides is the following:

    1. singletons
    2. providers
    3. factories

"""
from ._methods import factory, provider, singleton

__all__ = ['singleton', 'factory', 'provider']



from typing import Any, Type

import pytest

from antidote import Factory, factory, Tag, Wiring, world
from antidote._providers import (FactoryProvider, LazyProvider, ServiceProvider,
                                 TagProvider)


@pytest.fixture(autouse=True)
def test_world():
    with world.test.empty():
        world.provider(ServiceProvider)
        world.provider(FactoryProvider)
        world.provider(LazyProvider)
        world.provider(TagProvider)
        yield


class A:
    def __init__(self, **kwargs):
        self.kwargs = kwargs


class B:
    pass


class C:
    pass


def dummy() -> A:
    pass


@pytest.fixture(params=['function', 'class'])
def build(test_world, request):
    if request.param == 'function':
        @factory
        def build(**kwargs) -> A:
            return A(**kwargs)

        return build
    else:
        class ServiceFactory(Factory):
            def __call__(self, **kwargs) -> A:
                return A(**kwargs)

        return ServiceFactory


def test_simple(build: Type[Factory]):
    # working dependency
    assert isinstance(world.get(A @ build), A)
    assert world.get(A @ build) is world.get(A @ build)


def test_pass_through(build):
    if isinstance(build, type) and issubclass(build, Factory):
        build = build()

    assert isinstance(build(), A)

    x = object()
    a = build(x=x)
    assert a.kwargs == dict(x=x)


def test_auto_wire():
    world.singletons.add(C, C())

    with world.test.clone(keep_singletons=True):
        @factory
        def build(x: C = None) -> A:
            return x

        assert world.get(A @ build) is world.get(C)

    with world.test.clone(keep_singletons=True):
        @factory(auto_wire=True)
        def build(x: C = None) -> A:
            return x

        assert world.get(A @ build) is world.get(C)

    with world.test.clone(keep_singletons=True):
        @factory(auto_wire=False)
        def build(x: C = None) -> A:
            return x

        assert world.get(A @ build) is None


def test_tags():
    tag = Tag()
    with world.test.clone():
        @factory(tags=[tag])
        def build() -> A:
            return A()

        a = next(world.get(tag).values())
        assert isinstance(a, A)
        assert a is world.get(A @ build)

    with world.test.clone():
        class ServiceFactory(Factory):
            __antidote__ = Factory.Conf(tags=[tag])

            def __call__(self) -> A:
                return A()

        a = next(world.get(tag).values())
        assert isinstance(a, A)
        assert a is world.get(A @ ServiceFactory)

    with world.test.empty():
        world.provider(FactoryProvider)

        with pytest.raises(RuntimeError):
            @factory(tags=[tag])
            def build() -> A:
                return A()

    with world.test.empty():
        world.provider(FactoryProvider)

        with pytest.raises(RuntimeError):
            class ServiceFactory(Factory):
                __antidote__ = Factory.Conf(tags=[tag])

                def __call__(self) -> A:
                    return A()


def test_invalid_dependency(build: Type[Factory]):
    with pytest.raises(ValueError, match="Unsupported dependency.*"):
        B @ build

    with pytest.raises(ValueError, match="Unsupported dependency.*"):
        B @ build.with_kwargs(x=1)


def test_with_kwargs(build: Type[Factory]):
    x = object()
    a = world.get(A @ build.with_kwargs(x=x))
    assert a.kwargs == dict(x=x)

    with pytest.raises(ValueError, match=".*with_kwargs.*"):
        A @ build.with_kwargs()


def test_invalid_conf():
    with pytest.raises(TypeError, match=".*__antidote__.*"):
        class Dummy(Factory):
            __antidote__ = 1

            def __call__(self) -> A:
                pass


def test_missing_call():
    with pytest.raises(TypeError, match="__call__"):
        class ServiceFactory(Factory):
            pass


def test_missing_return_type_hint():
    with pytest.raises(ValueError):
        @factory
        def faulty_service_provider():
            return A()

    with pytest.raises(TypeError):
        @factory
        def faulty_service_provider2() -> Any:
            return A()

    with pytest.raises(ValueError):
        class FaultyServiceFactory(Factory):
            def __call__(self):
                return A()

    with pytest.raises(TypeError):
        class FaultyServiceFactory2(Factory):
            def __call__(self) -> Any:
                return A()


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(f=1), pytest.raises(TypeError, match=".*function.*")),
    (dict(f=dummy, auto_wire=1), pytest.raises(TypeError, match=".*auto_wire.*")),
])
def test_invalid_args(kwargs, expectation):
    with expectation:
        factory(kwargs.pop('f'), **kwargs)


def test_no_subclass_of_service():
    class Dummy(Factory):
        def __call__(self, *args, **kwargs) -> A:
            pass

    with pytest.raises(TypeError, match=".*abstract.*"):
        class SubDummy(Dummy):
            def __call__(self, *args, **kwargs) -> B:
                pass


@pytest.mark.parametrize('kwargs, expectation', [
    (dict(public=object()), pytest.raises(TypeError, match=".*public.*")),
    (dict(singleton=object()), pytest.raises(TypeError, match=".*singleton.*")),
    (dict(tags=object()), pytest.raises(TypeError, match=".*tags.*")),
    (dict(tags=['dummy']), pytest.raises(TypeError, match=".*tags.*")),
    (dict(wiring=object()), pytest.raises(TypeError, match=".*wiring.*")),
])
def test_conf_error(kwargs, expectation):
    with expectation:
        Factory.Conf(**kwargs)


@pytest.mark.parametrize('kwargs', [
    dict(singleton=False),
    dict(tags=(Tag(),)),
    dict(public=True),
    dict(wiring=Wiring(methods=['method'])),
])
def test_conf_copy(kwargs):
    conf = Factory.Conf(singleton=True, tags=[], public=False).copy(**kwargs)
    for k, v in kwargs.items():
        assert getattr(conf, k) == v
